import os.path as path

from clera import *
from configparser import ConfigParser

def editor():
    DEFAULT_CONFIG = ConfigParser()

    DEFAULT_CONFIG['MAIN'] = {
        'font_family' : "monospace",
        'size' : '16',
    }


    DIR_NAME = f'{path.dirname(__file__)}'

    def init_config_file():
        if path.isfile(f'{DIR_NAME}/config.ini') == False:
            with open(f'{DIR_NAME}/config.ini', 'w') as config_file:
                DEFAULT_CONFIG.write(config_file)

    init_config_file()

    class Settings:
        def __init__(self, filename):
            self.config = ConfigParser()
            self.filename = filename
            self.config_dict = dict()

        def scan(self):
            self.config.read(self.filename)

            for section in self.config.sections():
                self.config_dict[section] = dict(self.config.items(section))


        def get(self, name):
            self.scan()
            section = self.config_dict[name]

            self.make_python(section)        
            
            return section
        
        def add(self):
            return self.config

        
        def update(self):
            for KEY in self.config_dict:
                VALUE = self.config_dict[KEY]
                self.make_config(VALUE)

                self.config[KEY] = VALUE

                with open(self.filename, 'w') as file:
                    self.config.write(file)

        def make_python(self, section):
            for KEY in section:
                VALUE = section[KEY]
                if VALUE == 'True':
                    section[KEY] = True
                elif VALUE == 'False':
                    section[KEY] = False
                elif VALUE == 'None':
                    section[KEY] = None
        
        def make_config(self, section):
            for KEY in section:
                VALUE = section[KEY]
                if VALUE == None:
                    section[KEY] = 'None'

    SETTINGS = Settings(f"{DIR_NAME}/config.ini")
    MAIN = SETTINGS.get("MAIN")



    # Automatic indention
    lines = [0]
    def make_line_width():
        line_width = '40px'

        if int(MAIN['size']) > 16:
            line_width = '55px'
        elif int(MAIN['size']) > 14:
            line_width = '45px'

        return line_width

    def set_line_number(line_widget, text_widget):
        # area = GET('line')
        line_widget.value(''.join([
            f'{line + 1}\n' 
            for line in range(
                len(str(text_widget).split('\n'))
                )
            ]).removesuffix('\n'))
        line_widget.select_all()
        line_widget.alignment('right')
        line_widget.setcursor(line_widget.cursor())

        if len(str(text_widget).split('\n')) >= 10000:
            line_widget.style('max-width: 65px')
        elif len(str(text_widget).split('\n')) >= 1000:
            line_widget.style('max-width: 55px')
        else:
            line_widget.style(f'max-width: {make_line_width()}')

            

    def auto_indent():
        txt = GET('txt')
        txt_value = str(txt).split('\n')

        line = lines[0]
        curr_line = len(txt_value)
        def tab(n: None = None):
            c, p = txt.cursor()

            first = str(txt)[:p]
            last = str(txt)[p:]

            txt_value = first.split('\n')
            
            if n == 'add':
                txt_value[-1] += '}'
            else:
                txt_value[-1] = '\t'
                p += len('\t')


            last = last.split('\n')

            if last[0] == '':
                last.pop(0)

            if n == 'close':
                txt_value.append('}')
            
            txt_value += last
            
            # if txt_value[-1] == '':
            #     txt_value.pop(-1)
            
            txt_value = '\n'.join(txt_value)
            value = txt_value
            
            cursor = c, p

            txt.value(value)

            txt.setcursor(cursor)

        
        ln = txt_value[-1]
        c, p = txt.cursor()

        if str(txt)[:p].split('\n')[-1] == '':
            if line < curr_line:
                end = str(txt)[:p].strip()

                try:
                    end = end[-1].split()[0]
                except:
                    ...

                if  end == ';':
                    tab()
                elif end == '{':
                    try:
                        after = str(txt)[p:].strip()[0]
                    except:
                        after = None

                        if after != '}':
                            tab('close')
                
                lines[0] = curr_line
            else:
                lines[0] = 0

        elif ln == '\t}':
            txt_value[-1] = txt_value[-1].replace('\t', '')
            txt_value = '\n'.join(txt_value)
            cursor, postion = txt.cursor()
            postion -= len('\t')
            cursor = cursor, postion
            txt.value(txt_value)

            txt.setcursor(cursor)
        else:
            ...

        line_w = GET('line')
        txt_w = GET('txt')
        

        if len(str(line_w).split('\n')) < len(str(txt_w).split('\n')) or len(str(line_w).split('\n')) > len(str(txt_w).split('\n')):
            set_line_number(line_w, txt)

    #interface
    window = Window(content_margin=0, frame=False, movable=True, fixed_size=(420, 270), spacing=0)
    line_width = make_line_width()

    settings_styling = f'''
        font-size: {MAIN['size']}px;
        font-family: {MAIN['font_family']};
    '''

    css = f'''
    button [
        border: 0px solid;
        padding: 3px 2px;
    ]

    select [
        height: 25px;
    ]

    button:hover [
        background: rgba(131, 130, 130, 0.5);
        border-radius: 5px;
    ]

    scrollbar:vertical [
        background: rgba(0, 0, 0, 0);
        width: 15px;
        margin: 0;
        max-width: 5px;
    ]

    scrollbar::handle:vertical [
        background: gray;   
    ]

    scrollbar::add-line:vertical [
        height: 0px;
    ]

    scrollbar::sub-line:vertical [
        height: 0px;
    ]

    line [
        max-width: {line_width};
        border: 0px solid;
        background: rgb(19, 19, 19);
        color: rgb(131, 130, 130);
        {settings_styling}
    ]

    txt [
        {settings_styling}
    ]

    scrollbar:horizontal [
        background: rgba(0, 0, 0, 0);
        margin: 0;
        max-height: 5px;
    ]

    scrollbar::handle:horizontal [
        background: gray;   
    ]

    scrollbar::add-line:horizontal [
        height: 0px;
    ]

    scrollbar::sub-line:horizontal [
        height: 0px;
    ]
    '''
    # line {
    #     max-width: 65px;
    #     border: 0px solid;
    #     background: rgb(19, 19, 19);
    #     color: rgb(131, 130, 130);
    # }


    POPUP_STYLING = '''
    button { 
        padding: 5px 20px 8px 20px; 
        background: rgba(131, 130, 130, 0.5); 
        border-radius: 5px;
    } 

    button:hover {
        background: rgba(131, 130, 130, 1); 
        border-radius: 5px;
    }'''



    allow = '(Clera Style: *.cx)'


    editor_state = [False]
    file_path = [None]




    def handle(action):
        textarea = GET('txt')

        def save_as():
            file, format = File.save(filter=allow)
            file_path[0] = file
            if file.strip().endswith('cx'):
                ...
            else:
                file = f'{file.strip()}.cx'

            if len(file) != 0:
                with open(file, 'w') as file:
                    file.write(str(textarea))
                    editor_state[0] = False
                    GET('line').style('border-right: 0px;')
                    return True
            else:
                return False


        def handle_open():
            file, format = File.open(filter=allow)
            file_path[0] = file

            
            with open(file, 'r') as file:
                file = file.read()
                textarea.value(file)
                editor_state[0] = False
                GET('line').style('border-right: 0px;')

        
        
        def make_new():
            file_path[0] = None
            textarea.clear()
            editor_state[0] = False
            GET('line').style('border-right: 0px;')


        def unsaved(action):
            def cancel():
                GET('warning-popup').delete()

            def dont_save():
                cancel()

                if action == 'open':
                    handle_open()
                elif action == 'new':
                    make_new()
                elif action == 'exit':
                    window.quit()

            def save():
                cancel()
                query = handle_save()
                
                if query == True:
                    if action == 'open':
                        handle_open()
                    elif action == 'new':
                        make_new()
                    elif action == 'exit':
                        window.quit()
                else:
                    if action == 'open':
                        handle_open()
                    elif action == 'exit':
                        window.quit()

            warning = [
                [Text('Do you want to Save Document?')],
                [Button('Save', save, id='popup-save'), Button("Don't Save", dont_save, id='popup-d-save'), Button('Cancel', cancel,id='popup-cancel', shortcut='Esc')]
            ]

            Popup('Notification', warning, modal=True, id='warning-popup', lock=True)
            GET('popup-save').style(POPUP_STYLING)
            GET('popup-d-save').style(POPUP_STYLING)
            GET('popup-cancel').style(POPUP_STYLING)


        
        def handle_save():
            if file_path[0] != None and len(file_path[0])  != 0:
                with open(file_path[0], 'w') as file:
                    file.write(str(textarea))
                    editor_state[0] = False
                    GET('line').style('border-right: 0px;')
            else:
                save_as()
            

        if action == 'open':
            if editor_state[0] == True:
                unsaved('open')
            else:
                handle_open()
        elif action == 'save':
            handle_save()
        elif action  == 'save_as':
            save_as()
        elif action == 'new':
            if editor_state[0] == True:
                unsaved('new')
            else:
                make_new()

            
            GET('txt').focus()
        elif action == 'exit':
            if editor_state[0] == True:
                unsaved('exit')
            else:
                window.quit()
        else:
            ...

        GET('txt').focus()
        
        


    def changed():
        auto_indent()

        if editor_state[0] == False:
            editor_state[0] = True
            GET('line').style('border-right: 1px solid gray;')
        else:
            ...

    # end

    state = ['on']

    def switch():
        current_state = state[0]
        _switch = GET('switch')
        _textarea = GET('txt')

        if current_state == 'on':
            _switch.icon('icons/off.png')
            state[0] = 'off'
            _textarea.disable()
        elif current_state == 'off':
            _switch.icon('icons/on.png')
            state[0] = 'on'
            _textarea.enable()

    def settings():
        def pop_action(action):
            popup = GET('settings')

            if action == 'cancel':
                popup.delete()
                GET('txt').focus()

            elif action == 'apply':
                size_value = GET('-size-').current()
                font = GET('font').current()

                MAIN['size'] = size_value
                MAIN['font_family'] = font
                SETTINGS.update()

                settings_styling = f'''
                    font-size: {MAIN['size']}px;
                    font-family: {MAIN['font_family']};
                '''

                GET('line').style(f'{settings_styling};max-width: {make_line_width()}')
                GET('txt').style(settings_styling)

        size = [
            option(12),
            option(13),
            option(14),
            option(15),
            option(16),
            option(17),
            option(18),
        ]
        
        font_options = [
            option('monospace'),
            option('courier'),
            option('arial'),
            option('comic sans ms'),
            option('helvatica'),
        ]


    # Input('font family', sizepolicy=fixed, value=MAIN['font_family'], id='font')
        settings = [
            [Select(font_options, id='font'), Select(size, placeholder=MAIN['size'], id='-size-')],
            [Button('Apply', call(pop_action, 'apply'), id='apply'), Button('Cancel', call(pop_action, 'cancel') , id='cancel')]
        ]

        # settings = [
        #     [Text('UNAVAILABLE!', alignment=center)],
        #     [Button('Cancel', call(pop_action, 'cancel') , id='cancel')]
        # ]
        
        popup(id='settings', widgets=settings,frame=False, lock=True, size=(230, 100), modal=True)
        GET('apply').style(POPUP_STYLING)
        GET('cancel').style(POPUP_STYLING)
        GET('-size-').style('max-width: 40px;')

        sizes = [12, 13, 14, 15, 16, 17, 18]
        GET('-size-').set(sizes.index(int(MAIN['size'])))

        fonts = ['monospace', 'courier', 'arial', 'comic sans ms', 'helvatica']
        GET('font').set(fonts.index(MAIN['font_family']))

    tool_items = [
        Button(func=call(handle, 'new'), icon=f'{DIR_NAME}/../icons/add-file.png', icon_size=25, tooltip='new', shortcut='Ctrl+N'),
        Button(func=call(handle, 'open'), icon=f'{DIR_NAME}/../icons/documents.png', icon_size=25, tooltip='open', shortcut='Ctrl+O'),
        Button(func=call(handle, 'save'), icon=f'{DIR_NAME}/../icons/save.png', icon_size=25, tooltip='save', shortcut='Ctrl+S'),
        Button(func=call(handle, 'save_as'), icon=f'{DIR_NAME}/../icons/save-as.png', icon_size=25, tooltip='save as', shortcut='Ctrl+Shift+S'),
        Button(id='switch', func=switch, icon=f'{DIR_NAME}/../icons/on.png', icon_size=25, tooltip='switch', shortcut='F5'),
        Button(id='gear', func=settings, icon=f'{DIR_NAME}/../icons/gear.png', icon_size=25, tooltip='settings'),
        Text(id='space', sizepolicy=('expand', 'expand')),
        Button(func=call(handle, 'exit'), icon=f'{DIR_NAME}/../icons/close.png', icon_size=25, tooltip='exit', shortcut='Esc'),
    ]

    toolbar('main', tool_items, position='right', orientation='v', id='tools', border=False)
    # Textarea('line', disabled=True, value='1'), 

    Box([[Textarea('line', disabled=True, value='1'), Textarea('txt', text_changed=changed, wordwrap=False)]])

    GET('txt').scrollbar(GET('line').scrollbar(id='scroll'))
    GET('line').alignment('right')

    keywords = [
        'button', 
        'text', 
        'input', 
        'checkbox', 
        'radiobutton',
        'toolbar', 
        'fieldset', 
        'item', 
        'textarea', 
        'window',
        'listwidget', 
        'select', 
        'progressbar', 
        'slider', 
        'dial',
        'tabwidget', 
        'popup', 
        'group',
        'stack',
        'scrollbar'
    ]


    synthax = {
        r'(.*).*\{': {
            'color': 'rgb(14, 106, 182)'
        },
        
        r(keywords): {
            'color': 'red',
        }, 
        
        r'(.*).*:': {
            'color': 'rgb(131, 130, 130)'
        },

        '\'(.*?)\'|\"(.*?)\"': {
            'color': 'yellow',
        },

        ':(.*);|:(.*)': {
            'color': 'red'
        },

        r'(-)[0-9]': {
            'color': 'rgb(14, 106, 182)'
        },

        r'[0-9](.*)[0-9]': {
            'color': 'rgb(14, 106, 182)'
        },

        r'px': {
            'color': 'rgb(131, 130, 130)'
        },

        r(['rgb',"rgba"]): {
            'color': 'rgb(155, 102, 197)'
        },

        
        r'\:|\;|\,': {
            'color': 'white'
        },

        r'[0-9]': {
            'color': 'rgb(14, 106, 182)'
        },

        

        r'\A#(.*)': {
            'color': 'rgb(94, 134, 0)'
        },
        
        r'\{|\}': {
            'color': 'rgb(235, 211, 3)'
        },
        
        r'\(|\)': {
            'color': 'rgb(112, 66, 150)'
        },
        
        
    }


    Highlight('txt', synthax)
    GET('txt').focus()
    window.run(css)



class SizePolicy:
    expand          : ...
    fixed           : ...

class Orientation:
    vertical        : ...
    horizontal      : ...

class Alignment:
    center          : ...
    vcenter         : ...
    hcenter         : ...
    justify         : ...
    top             : ...
    bottom          : ...
    left            : ...
    right           : ...
    
class Position:
    top             : ...
    bottom          : ...
    left            : ...
    right           : ...

class WidgetType:
    normal          : ...
    static          : ...

class InputType:
    standard        : ...
    password        : ...
    noecho          : ...

class SelectionMode:
    extended        : ...
    noselection     : ...
    multi           : ...
    single          : ...

class WindowStyle:
    windows         : ...
    fusion          : ...
    breeze          : ...
    oxygen          : ...
    curve           : ...

class ButtonStyle:
    circle          : ...
    square          : ...

class OpenFileType:
    single          : ...
    multi           : ...

def call(func: function, *args, **kwargs): ...
def r(list: list) -> str: ...
def editor(): ...


from PySide6.QtCore import QSize, QPoint, QEvent
from PySide6.QtGui import QIcon, Qt, QMouseEvent, QScreen, QSyntaxHighlighter
from PySide6.QtWidgets import QMainWindow, QToolBar, QStatusBar, QGroupBox, QButtonGroup
from PySide6.QtWidgets import QWidget,  QHBoxLayout, QVBoxLayout,  QGridLayout, QTabWidget, QMenu
from PySide6.QtWidgets import QDialog, QFileDialog, QStyleFactory, QStackedWidget
################################
################################
from .utils import *
from .widgets import *

__all__ = [
    'Window', 'Box', 'Grid', 'fieldset', 'Column', 'column', 'Group', 'group', 'Toolbar', 'toolbar',
    'Menubar', 'menubar', 'Statusbar', 'statusbar', 'GET', 'get', 'Button', 'button', 'Input',
    'input', 'Text', 'text', 'Image', 'image', 'CheckBox', 'checkbox', 'RadioButton', 'radiobutton',
    'Textarea', 'textarea', 'item', 'empty', 'separator', 'Exit', 'exit', 'Copy', 'copy', 'Cut', 'cut',
    'Paste', 'paste', 'Undo', 'undo', 'Redo', 'redo', 'link', 'ListWidget', 'listwidget', 'ListItem',
    'tab', 'TabWidget', 'tabwidget', 'Select', 'select', 'option', 'ProgressBar', 'progressbar',
    'Slider', 'slider', 'Dial', 'dial', 'Popup', 'popup', 'call', 'Titlebar', 'titlebar', 'File', 'file',
    'Folder', 'folder', 'ScrollArea', 'scrollarea', 'minimize', 'maximize', 'close', 'title', 'Highlight',
    'r', 'Stack', 'expand', 'fixed', 'vertical', 'horizontal', 'top', 'bottom', 'left', 'right', 'center',
    'vcenter', 'hcenter', 'justify', 'static', 'normal', 'standard', 'password', 'noecho', 'extended',
    'noselection', 'multi', 'single', 'windows', 'fusion', 'curve', 'breeze', 'oxygen','circle', 'square'
]


app = start()

DEFAULT_WINDOW_LAYOUT = QWidget()


def init_window_css(css_string):
    custom, builtins = console_css(css_string)

    for items in custom:
        name, property = items
        widget = WIDGET_ID_SAFE[name]

        if name in WIDGET_ID_SAFE:
            for value in compr:
                value = value.split('::')
                _type = value[1]

                if _type in str(type(widget)):
                    break

            property = property.replace(TYPE_MARKER, _type)
            widget.setStyleSheet(property)

    app.setStyleSheet(builtins)


def link(href: None = None):
    if href != None and href.strip().endswith('.cx'):
        LINK_ITEMS_INCLUDED.append(href)


def init_linked_files():
    for LINK_ITEM in LINK_ITEMS_INCLUDED:
        file_path = path.abspath(LINK_ITEM)

        is_file = path.isfile(file_path)
        is_valid_path = path.exists(file_path)

        if is_file != False and is_valid_path != False:
            with open(file_path, 'r') as file:
                file_content = file.read()

            init_window_css(file_content)


class MainWindow(QMainWindow):
    def __init__(self, title, icon, size, fixed_size, geometry, style, frame, movable, win_top, spacing, content_margin, move):
        super().__init__()

        self._qwidget = DEFAULT_WINDOW_LAYOUT
        self._title = title
        self._size = size
        self._fixed_size = fixed_size
        self._icon = icon
        self._geometry = geometry
        self._movable = movable
        self._frame = frame
        self._top = win_top
        self._spacing = spacing
        self._content_margin = content_margin
        self._move = move

        self._query_fixed_size = False
        self._custom_title_bar = False

        width, height = self._size
        if width != None and height != None:
            self.resize(int(width), int(height))

        positionx, positiony = self._move
        if positionx != None and positiony != None:
            self.move(positionx, positiony)

        if self._geometry != None:
            if len(self._geometry) == 4:
                x, y, w, h = self._geometry
                self.setGeometry(x, y, w, h)

        self.setWindowTitle(self._title)

        if self._top == True:
            self.setWindowFlags(Qt.WindowStaysOnTopHint)

        app.setStyle(style)

        fixed_width, fixed_height = self._fixed_size
        if fixed_width != None and fixed_height != None:
            self._query_fixed_size = True
            self.setFixedSize(QSize(int(fixed_width), int(fixed_height)))

        global layout
        layout = QVBoxLayout()

        layout.setSpacing(self._spacing)

        if type(self._content_margin) != int:
            if len(self._content_margin) == 4:
                left, top, right, bottom = self._content_margin
                layout.setContentsMargins(left, top, right, bottom)
            elif len(self._content_margin) == 2:
                top_bottom, left_right = self._content_margin
                layout.setContentsMargins(
                    left_right, top_bottom, left_right, top_bottom)
            else:
                ...
        else:
            layout.setContentsMargins(
                self._content_margin, self._content_margin, self._content_margin, self._content_margin)

        self._qwidget.setLayout(layout)

        if self._frame == False:
            self.setWindowFlag(Qt.FramelessWindowHint)

        self.setWindowIcon(QIcon(get_path(self._icon)))

    def WindowMenubar(self, menu_item, actions: list | None = None):
        if menu_item != None:
            menu_bar = self.menuBar()
            Parent = menu_bar.addMenu(menu_item)

            if actions != None:
                for item in actions:

                    item_type = item[0]

                    if item_type == ELEM_TYPE_ITEM:

                        item_label = item[1]

                        if item_label != DEFAULT_SEPARATOR_MARKER:
                            Child = Parent.addAction(INIT_ITEM(self, item))
                        else:
                            Parent.addSeparator()
                    elif item_type == ELEM_TYPE_SEPARATOR:
                        Parent.addSeparator()

                    else:
                        pass
                        # DO NOTHING FOR NOW. SETUP LATER

    def WindowToolbar(self, name, tool_items, movable, position, ID, size, border, orientation, newline):

        toolbar = INIT_WIDGET(ID, QToolBar(name))

        if newline == True:
            self.addToolBarBreak()

        width, height = size
        if width != None or height != None:
            toolbar.setIconSize(QSize(width, height))

        toolbar_position = position.lower()

        positions = {
            top     : Qt.TopToolBarArea,
            bottom  : Qt.BottomToolBarArea,
            left    : Qt.LeftToolBarArea,
            right   : Qt.RightToolBarArea
        }

        try:
            self.addToolBar(positions[toolbar_position], toolbar)
        except:
            raise ValueError(position)

        if tool_items != None:
            for item in tool_items:
                if type(item) != type(list()):
                    widget_type, widget_items = check_func(item)
                    LayItOut(toolbar, widget_type, widget_items)
                    # toolbar.addWidget(widget)
                else:

                    item_type = item[0]

                    if item_type == ELEM_TYPE_ITEM:

                        item_label = item[1]

                        if item_label != DEFAULT_SEPARATOR_MARKER:
                            toolbar.addAction(INIT_ITEM(self, item))
                        else:
                            toolbar.addSeparator()
                    elif item_type == ELEM_TYPE_SEPARATOR:
                        toolbar.addSeparator()

                    else:
                        pass
                        # DO NOTHING FOR NOW. SETUP LATER

        toolbar.setMovable(movable)

        orientations = {
            vertical: Qt.Vertical,
            horizontal: Qt.Horizontal
        }

        try:
            if orientation.lower() in ['v', 'h']:
                change = {
                    'v': vertical,
                    'h': horizontal
                }

                orientation = change[orientation.lower()]

            toolbar.setOrientation(orientations[orientation.lower()])
        except:
            raise ValueError(orientation)

        if border == False:
            toolbar.setStyleSheet('border-bottom: 0px')

    # Status bar functions

    def init_status(self):
        self.setStatusBar(QStatusBar(self))

    def ADD_PERMANENT_WIDGET(self, widget_type, widget_items):
        lyt = self.statusBar()
        LayItOut([lyt, 'addPermanentWidget'], widget_type, widget_items)

    def ADD_NORMAL_WIDGET(self, widget_type, widget_items):
        lyt = self.statusBar()
        LayItOut([lyt, 'addWidget'], widget_type, widget_items)

    def SHOW_STATUSBAR_MESSAGE(self, text, time):
        if time != None:
            time = time * 1000
            self.statusBar().showMessage(text, time)
        else:
            self.statusBar().showMessage(text)

    def CLEAR_STATUSBAR_MESSAGE(self):
        self.statusBar().clearMessage()

    def REMOVE_STATUSBAR_WIDGET(self, ID):
        if ID in WIDGET_ID_SAFE:
            self.statusBar().removeWidget(WIDGET_ID_SAFE[ID])
        else:
            raise IdError(f'id "{ID}" does not exist')

    def mousePressEvent(self, event):
        if self._movable == True:
            window.setCursor(Qt.SizeAllCursor)

        self.old_position = event.globalPos()
        return self.old_position

    def mouseMoveEvent(self, event):
        try:
            omega = QPoint(event.globalPos() - self.old_position)

            if self._custom_title_bar == False and self._movable == True and self._frame == False:
                self.move(self.x() + omega.x(), self.y() + omega.y())
                self.old_position = event.globalPos()
            else:
                return omega
        except:
            ...
        
    def mouseReleaseEvent(self, event: QMouseEvent):
        window.setCursor(Qt.ArrowCursor)

    def WindowTitlebar(self, icon, title, widgets, alignment, backgound_color,  text_color, height,
                       button_style):
        current_window_style = window.style().name().lower()

        if current_window_style == SYSTEM_PLATFORM_FUSION:
            CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED = FUSION_CONTROLS()
        elif current_window_style == SYSTEM_PLATFORM_WINDOWS or 'windows' in current_window_style:
            CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED = WINDOWS_CONTROLS()
            if SYSTEM_OS_PLATFORM == 'windows':
                CONTROL_MAXIMIZED = '🗖'
        else:
            CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED = CLERA_CONTROLS()

        if self._custom_title_bar == False and self._frame == False:
            if alignment != None:
                if alignment.lower() == 'center':
                    align = 'center'
                else:
                    align = None
            else:
                align = None

            def window_state():
                state = str(self.windowState())
                state = state.split('.')
                state = state[1].lower()
                return state

            def control_action(key):
                state = window_state()
                max = GET(MAXIMIZE_BUTTON_ID)

                if key == 'max':
                    if state == 'windowmaximized':
                        self.setWindowState(Qt.WindowNoState)
                        max.value(CONTROL_MAXIMIZED)
                    elif state == 'windownostate':
                        self.setWindowState(Qt.WindowMaximized)
                        max.value(CONTROL_RESTORE)
                elif key == 'min':
                    self.setWindowState(Qt.WindowMinimized)
                    try:
                        max.value(CONTROL_MAXIMIZED)
                    except:
                        ...
            if icon != None:
                titlebar_icon = Image(icon, id=ICON_IMAGE_ID, size=20)
            else:
                titlebar_icon = empty()

            title_items = [
                titlebar_icon,
                Text(title, id=TITLE_TEXT_ID,
                     sizepolicy=('expand', 'fixed'), alignment=align),
                Button(CONTROL_MINIMIZED, id=MINIMIZE_BUTTON_ID,
                       func=call(control_action, 'min'), focus=False),
                Button(CONTROL_MAXIMIZED, id=MAXIMIZE_BUTTON_ID,
                       func=call(control_action, 'max'), focus=False),
                Button(CONTROL_CLOSE, id=CLOSE_BUTTON_ID,
                       func=Window.quit, focus=False)
            ]

            container = INIT_WIDGET(
                '-clera_titlebar_container-', QToolBar('Container'))
            titlebar = INIT_WIDGET('titlebar', QToolBar('Titlebar'))

            class init_titlebar(QWidget):
                def __init__(self):
                    super().__init__()
                    cs_layout = QHBoxLayout()
                    self.setLayout(cs_layout)
                    cs_layout.setContentsMargins(0, 0, 0, 0)
                    cs_layout.addWidget(titlebar)
                    window.setCentralWidget(self)

                def mousePressEvent(self, event: QMouseEvent):
                    self.old_position = window.mousePressEvent(event)
                    window.setCursor(Qt.SizeAllCursor)

                def mouseMoveEvent(self, event: QMouseEvent):
                    omega = window.mouseMoveEvent(event)
                    window.move(window.x() + omega.x(), window.y() + omega.y())
                    self.old_position = window.mousePressEvent(event)

                def mouseReleaseEvent(self, event: QMouseEvent):
                    window.setCursor(Qt.ArrowCursor)

            container_socket = init_titlebar()
            container_socket.setContentsMargins(0, 0, 0, 0)

            container.addWidget(container_socket)
            container.setStyleSheet('margin: 0; padding: 0; border: 0px;')

            container.setMovable(False)
            titlebar.setMovable(False)

            self.addToolBar(Qt.TopToolBarArea, container)

            container.toggleViewAction().setVisible(False)
            titlebar.toggleViewAction().setVisible(False)

            titlebar.setIconSize(QSize(20, 20))

            # self.addToolBar(Qt.TopToolBarArea, titlebar)

            self.addToolBarBreak()

            if widgets != None:
                if align == 'center':
                    for item in title_items:
                        append_items(widgets, item)
                    title_items = widgets
                else:
                    window_title = [title_items[0]]
                    controls = title_items[1:]

                    for item in widgets:
                        append_items(window_title, item)
                    else:
                        for item in controls:
                            append_items(window_title, item)

                    title_items = window_title

            for item in title_items:
                if type(item) != type(list()):
                    widget_type, widget_items = check_func(item)
                    LayItOut(titlebar, widget_type, widget_items)
                    # titlebar.addWidget(widget)

            titlebar.setOrientation(Qt.Horizontal)

            button_style_query = button_style.lower()

            button_styles = {
                circle: CONTROL_BUTTON_CIRCLE,
                square: CONTROL_BUTTON_SQUARE 
            }

            try:
                control_button_style = button_styles[button_style_query]
            except:
                ...
                # raise an error

            title = GET(TITLE_TEXT_ID)
            minimize = GET(MINIMIZE_BUTTON_ID)
            maximize = GET(MAXIMIZE_BUTTON_ID)
            close = GET(CLOSE_BUTTON_ID)

            def init_title_icon(icon, value: str = 'left'):
                if icon != None:
                    icon = GET(ICON_IMAGE_ID)
                    icon.style(f'margin-{value}: 10px;')

            init_title_icon(icon)

            if current_window_style == SYSTEM_PLATFORM_FUSION:
                fusion_style(minimize, maximize, close, control_button_style)
            elif current_window_style == SYSTEM_PLATFORM_WINDOWS or current_window_style == 'windowsvista':
                windows_style(minimize, maximize, close)
            else:
                ...

            title.style(
                f'color: {text_color}; padding: 7px 0; margin: 0 {height}px;')
            titlebar.setStyleSheet(
                f'border-bottom: 0px; background: {backgound_color};')
            self._custom_title_bar = True


# ------------------------------------------------------------------------#

###########################################################################
############################## WINDOWS CLASS ##############################
###########################################################################

# ------------------------------------------------------------------------#


class Window:
    def __init__(self, title: str = DEFAULT_WINDOW_TITLE, icon: str = DEFAULT_WINDOW_ICON, size: tuple = DEFAULT_WINDOW_SIZE,
                 geometry: tuple = DEFAULT_WINDOW_GEOMETRY, style: str = DEFAULT_WINDOW_STYLE, fixed_size: tuple = DEFAULT_WINDOW_FIXED_SIZE, frame: bool = True,
                 movable: bool = False, top: bool = False, spacing: int = 5, content_margin: tuple = (5, 5, 5, 5),
                 move: tuple = (None, None)):
        '''
        :param title:
        :param icon:
        :param size:
        :param geometry:
        :param style:
        :param fixed_size:
        :param frame:
        :param movable:
        :param top:
        :param spacing:
        :param content_margin:
        :param move:
        '''
        global window

        self.window_title = title
        self.window_icon = icon
        self.window_size = size
        self.window_fixed_size = fixed_size
        self.window_geometry = geometry
        self.window_style = style
        self.frame = frame
        self.movable = movable
        self.top = top
        self.spacing = spacing
        self.content_margin = content_margin
        self.move = move

        window = MainWindow(self.window_title, self.window_icon,
                            self.window_size, self.window_fixed_size, self.window_geometry, self.window_style, self.frame,
                            self.movable, self.top, self.spacing, self.content_margin, self.move)

    def run(self, css: None = None):
        window.show()
        if css != None:
            init_window_css(css)

        if len(LINK_ITEMS_INCLUDED) != 0:
            init_linked_files()

        def window_state():
            state = str(window.windowState())
            state = state.split('.')
            state = state[1].lower()
            return state

        state = window_state()
        try:
            max = GET(MAXIMIZE_BUTTON_ID)

            if state == 'windowmaximized':
                max.value(CONTROL_RESTORE)

            if window._query_fixed_size == True:
                max.delete()

            if window._custom_title_bar == True:
                menu_bar = window.menuBar()
                menu_bar.clear()

        except:
            ...

        app.exec()
        self.quit()

    def quit(self):
        app.quit()

    def update(self, remove_id, widget):
        replace_widget = WIDGET_ID_SAFE[remove_id]

        widget_type, widget_items = check_func(widget)

        try:
            layout.indexOf(replace_widget)
            LayItOut([layout, replace_widget, 'replaceWidget'],
                     widget_type, widget_items)
            layout.removeWidget(replace_widget)
            replace_widget.deleteLater()
            replace_widget = None
        except:
            pass

    def normal(self):
        window.setWindowState(Qt.WindowNoState)

    def minimize(self):
        window.setWindowState(Qt.WindowMinimized)

    def maximize(self):
        window.setWindowState(Qt.WindowMaximized)

    def details(self):
        screen = window.screen().virtualSize()
        _title = window.windowTitle()

        if len(_title) == 0:
            _title = None

        size = window.size()

        window_details = {
            'position': window.geometry().getCoords()[:2],
            'screen': (screen.width(), screen.height()),
            'style': window.style().name(),
            'size': (size.width(), size.height()),
            'title': _title,
            'system_styles': QStyleFactory().keys()
        }

        return window_details

    def title(self, value: any = None):
        if value != None:
            window.setWindowTitle(value)

    def _move(self, position: tuple = (None, None), top: bool = None):
        positionx, positiony = position

        if positionx != None and positiony != None:
            window.move(positionx, positiony)

    # def frames(self, value):
    #     # if value != window.frame:

    #     if value == False:
    #         window.setWindowFlag(Qt.FramelessWindowHint)
    #     elif value == True:
    #         window.setWindowFlag(~Qt.FramelessWindowHint)

    #     print(window.windowFlags())
    #     window.show()


# ------------------------------------------------------------------------#

###########################################################################
############################## CORE HANDLERS ##############################
###########################################################################

# ------------------------------------------------------------------------#


def handle_group(widgets, grouplayout):
    groupcontent = QButtonGroup(grouplayout)
    for items in widgets:

        WidgetType, items = check_func(items)
        groupcontent.addButton(
            LayItOut(grouplayout, WidgetType, items))

    groupcontent.setExclusive(True)


def init_column(hbox, vbox, property):
    hbox.addLayout(vbox)

    for properties in property[1]:
        col_hbox = QHBoxLayout()
        vbox.addLayout(col_hbox)
        for property in properties:
            WidgetType, property = check_func(property)
            if WidgetType != ELEM_TYPE_COLUMN:
                LayItOut(col_hbox, WidgetType, property)
            else:
                hbox.addLayout(vbox)
                for properties in property[1]:
                    col_hbox = QHBoxLayout()
                    vbox.addLayout(col_hbox)
                    for property in properties:
                        WidgetType, property = check_func(property)
                        LayItOut(col_hbox, WidgetType, property)


def init_menubar(menu_items):
    if menu_items != None:
        for items in menu_items:
            if len(items) == 2:
                name, actions = items
            else:
                if type(items[0]) == type(str()):
                    name = items[0]
                    actions = None
                else:
                    name = None
                    actions = items[0]

            window.WindowMenubar(name, actions)


# ------------------------------------------------------------------------#

###########################################################################
############################# LAYOUT PROCESSOR ############################
###########################################################################

# ------------------------------------------------------------------------#


def Box(widgets):
    containers = [ELEM_TYPE_COLUMN, ELEM_TYPE_FIELDSET,
                  ELEM_TYPE_GROUP]
    box_layout = QVBoxLayout()
    box_widget = QWidget()

    if get_parent(BOX_IDENTIFIER):
        box_widget = DEFAULT_WINDOW_LAYOUT
        box_layout = layout
        window.setCentralWidget(box_widget)

    box_widget.setLayout(box_layout)

    for items in widgets:
        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        box_layout.addLayout(hbox)

        for item in items:
            widget_type, widget_items = check_func(item)

            if widget_type not in containers:
                LayItOut(hbox, widget_type, widget_items)
            elif widget_type == ELEM_TYPE_COLUMN:
                init_column(hbox, vbox, widget_items)
            elif widget_type == ELEM_TYPE_FIELDSET:

                Notrequired, name, widgets, id, fieldset_layout = widget_items
                if widgets != None:

                    groupbox = INIT_WIDGET(id, QGroupBox(name))
                    hbox.addWidget(groupbox)

                    if fieldset_layout.lower() in DEFAULT_VERTICAL_TYPES:
                        groupbox_layout = QVBoxLayout()
                    elif fieldset_layout.lower() in DEFAULT_HORIZONTAL_TYPES:
                        groupbox_layout = QHBoxLayout()
                    else:
                        raise ValueError(fieldset_layout)

                    groupbox.setLayout(groupbox_layout)

                    for items in widgets:
                        widget_type, widget_items = check_func(items)
                        if widget_type == ELEM_TYPE_GROUP:

                            NotRequired, widgets, group_layout = items
                            if widgets != None:
                                if group_layout.lower() in DEFAULT_HORIZONTAL_TYPES:
                                    group_element_layout = QHBoxLayout()
                                elif group_layout.lower() in DEFAULT_VERTICAL_TYPES:
                                    group_element_layout = QVBoxLayout()
                                else:
                                    raise ValueError(group_layout)

                                groupbox_layout.addLayout(group_element_layout)
                                handle_group(widgets, group_element_layout)
                        else:
                            LayItOut(groupbox_layout,
                                     widget_type, widget_items)

            elif widget_type == ELEM_TYPE_GROUP:

                Notrequired, widgets, group_layout = widget_items

                if widgets != None:

                    if group_layout.lower() in DEFAULT_VERTICAL_TYPES:
                        group_element_layout = QVBoxLayout()
                    elif group_layout.lower() in DEFAULT_HORIZONTAL_TYPES:
                        group_element_layout = QHBoxLayout()
                    else:
                        raise ValueError(group_layout)
                    layout.addLayout(group_element_layout)
                    handle_group(widgets, group_element_layout)
    # returns a QWidget and layout like QVBoxLayout
    return [box_widget, box_layout]


def Grid(widgets):
    grid_widget = QWidget()
    grid = QGridLayout()

    if get_parent(GRID_IDENTIFIER):
        grid_widget = DEFAULT_WINDOW_LAYOUT
        grid_layout = layout

        window.setCentralWidget(grid_widget)
        grid_layout.addLayout(grid)

    grid_widget.setLayout(grid)

    grid_pos_x = 0
    grid_pos_y = 0

    for items in widgets:
        for item in items:
            widget_type, widget_items = check_func(item)
            LayItOut(grid, widget_type, widget_items, grid_pos_x, grid_pos_y)
            grid_pos_y += 1
        grid_pos_y = 0
        grid_pos_x += 1

    return [grid_widget, grid_layout]


class ScrollArea:
    def __init__(self, widgets: None = None, id: None = None, contain: bool = True):
        '''
        ScrollArea

        :param widgets:
        :param id:
        :param contain:
        '''

        elem_type = ELEM_TYPE_SCROLL_AREA
        widgets = widgets[0]
        self.scroll_widget = QWidget()

        if ELEM_TYPE_TAB not in get_parent() and SCROLL_AREA_IDENTIFIER not in get_parent():
            # layout.addWidget(scroll_area)
            window.setCentralWidget(self.scroll_widget)

        scroll_layout = QVBoxLayout()

        scroll_area = INIT_WIDGET(id, QScrollArea())
        scroll_layout.addWidget(scroll_area)
        self.scroll_widget.setLayout(scroll_layout)
        scroll_area.setWidget(widgets)
        scroll_area.setAlignment(Qt.AlignBottom)
        scroll_area.setWidgetResizable(contain)

        

    def __repr__(self):
        return [self.scroll_widget]

    def __call__(self):
        return [self.scroll_widget]


class scrollarea(ScrollArea):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Stack:
    def __init__(self, widgets: None = None, id: None = None, current_changed: None = None, widget_removed: None = None):
        '''
        Stack Widget

        :param widgets:
        :param id:
        :param current_changed:
        :param widget_removed:
        '''
        self.stacked_widget = INIT_WIDGET(id, QStackedWidget())
        
        def init_stack(widget):
            widget = widget[0]
            self.stacked_widget.addWidget(widget)

        if widgets != None:
            if process_request(ELEM_TYPE_SCROLL_AREA, str(type(widgets)).lower()):
                widgets = widgets()
            
            if type(widgets[0]) == sample_list:
                for widget in widgets:
                    init_stack(widget)    
            else:
                init_stack(widgets)    
        
        layout.addWidget(self.stacked_widget)
        DEFAULT_WINDOW_LAYOUT.setLayout(layout)

        self.stacked_widget.currentChanged.connect(current_changed)
        self.stacked_widget.widgetRemoved.connect(widget_removed)
        
        window.setCentralWidget(DEFAULT_WINDOW_LAYOUT)
    def set(self, index):
        self.stacked_widget.setCurrentIndex(index)

    def show(self, widget):
        widget, layout = widget
        self.stacked_widget.setCurrentWidget(widget)


class stack(Stack):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
############################## CORE ELEMENTS ##############################
###########################################################################

# ------------------------------------------------------------------------#


class Menubar:
    def __init__(self, menu_items: list | None = None):
        '''
        Menu Bar

        :param menu_items:
        '''

        self.menu_items = menu_items

        init_menubar(self.menu_items)


class menubar(Menubar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Titlebar:
    def __init__(self, icon: str = None, title: str = path.basename(sys.argv[0]), widgets: list = None, alignment: None = None, text_color: str = 'white',
                 background_color: str = 'rgb(32, 32, 32)', height: any = CUSTOM_TITLEBAR_HEIGHT, button_style: None = DEFAULT_BUTTON_STYLE):
        '''
        Titlebar

        :param title:
        :param widgets:
        :param alignment:
        :param text_color:
        :param background_color:
        :param height:
        :param button_style:
        '''
        window.WindowTitlebar(icon, title, widgets, alignment,
                              background_color, text_color, height, button_style)


class titlebar(Titlebar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Toolbar:
    def __init__(self, name, tool_items: list | None = None, movable: bool = False,
                 position: str = top, id: None = None, iconsize: tuple = (None, None),
                 border: bool = True, orientation: str = SET_HORIZONTAL, newline: bool = False):
        '''
        Toolbar Element

        :param name:
        :param tool_items:
        :param movable:
        :param position:
        :param id:
        :param iconsize:
        :param border:
        :param orientation:
        :param newline:
        '''

        window.WindowToolbar(name,  tool_items, movable,  position,  id,  iconsize,  border,
                             orientation, newline)


class toolbar(Toolbar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Statusbar:
    def __init__(self):
        window.init_status()

    def message(self, text: str = '', time: None = None):
        '''
        Display temporary status message

        :param text:
        :param time: in seconds.
        '''
        window.SHOW_STATUSBAR_MESSAGE(text, time)

    def clear(self):
        window.CLEAR_STATUSBAR_MESSAGE()

    def add(self, widget, type: str = SET_NORMAL):
        type = type.upper()
        widget_type, widget_items = check_func(widget)
        if type == SET_NORMAL:
            window.ADD_NORMAL_WIDGET(widget_type, widget_items)
        elif type == SET_STATIC:
            window.ADD_PERMANENT_WIDGET(widget_type, widget_items)
        else:
            raise ValueError(f'"{type}" is an invalid type value')

    def remove(self, id):
        window.REMOVE_STATUSBAR_WIDGET(id)


class statusbar(Statusbar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class File(QFileDialog):
    def __init__(self):
        ...

    def open(self: None = QFileDialog, caption: str = None, filter: str = '(All Files: *)', directory: str = None, type: str = DEFAULT_OPEN_FILE_TYPE):
        '''
        Open File Dialog

        :param caption:
        :param filter:
        :param directory:
        :param type:
        '''

        open_type = type.lower()

        if type == OPEN_FILE_TYPE_SINGLE:
            file = self.getOpenFileName(
                caption=caption,
                dir=directory,
                filter=get_filter(filter))

            # file_names = file[0]
            return file

        elif open_type == OPEN_FILE_TYPE_MULTI:
            files = self.getOpenFileNames(
                caption=caption,
                dir=directory,
                filter=get_filter(filter))

            return files

    def save(self: None = QFileDialog, filter: str = '(All Files: *)'):
        '''
        Save File Dialog

        :param filter:
        '''

        file = self.getSaveFileName(filter=get_filter(filter))

        return file


class file(File):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Folder(QFileDialog):
    def __init__(self, caption: str = None, directory: str = None):
        '''
        Folder Dialog

        :param caption:
        :param directory:
        '''
        
        self.folder = self.getExistingDirectory(caption=caption,
                                                dir=directory)

    def __repr__(self):
        return self.folder


class folder(Folder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Popup:
    def __init__(self, title: None = None, widgets: None = None, size: tuple = (None, None), id: None = None,
                 modal: bool = False, frame: bool = True, lock: bool = False):
        '''
        Popup Window

        :param title:
        :param widgets:
        :param size:
        :param id:
        :param modal:
        :param frame:
        :param lock:
        '''

        if lock == True:
            parent = DEFAULT_WINDOW_LAYOUT
        else:
            parent = None

        elem_type = ELEM_TYPE_POPUP
        popup = INIT_WIDGET(id, QDialog(parent))

        self.popup = popup

        popup.setWindowTitle(title)

        if widgets != None:
            popup_layout = Box(widgets)

            popup.setLayout(popup_layout[1])

        popup.setModal(modal)

        width, height = size
        if width != None and height != None:
            popup.resize(int(width), int(height))

        # popup.resize(200, 120)

        if frame == False:
            popup.setWindowFlag(Qt.FramelessWindowHint)

        if popup.isModal:
            popup.show()
        else:
            popup.exec()

        self.rtn = [elem_type, id, popup]

    def __call__(self):
        return self.rtn

    def result(self):
        return self.popup.result()
    

class popup(Popup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TabWidget:
    def __init__(self, tabs: None = None, id: None = None, movable: bool = False, closable: bool = False,
                 close_requested: None = None, clicked: None = None):
        '''
        Tab Widget

        :param tabs:
        :param id:
        :param movable:
        :param closable:
        :param close_requested:
        :param clicked:
        '''

        elem_type = ELEM_TYPE_TAB

        tab_widget = INIT_WIDGET(id, QTabWidget())

        if tabs != None:
            add_tabs(tab_widget, tabs)

        tab_widget.setMovable(movable)
        tab_widget.setTabsClosable(closable)

        tab_widget.tabCloseRequested.connect(close_requested)
        tab_widget.tabBarClicked.connect(clicked)

        layout.addWidget(tab_widget)
        window.setCentralWidget(DEFAULT_WINDOW_LAYOUT)

        self.rtn = [elem_type, id]

    def __call__(self):
        return self.rtn


class tabwidget(TabWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Highlighter(QSyntaxHighlighter):
    def __init__(self, mapping, parent=None):
        super().__init__(parent)
        self.mapping = mapping

    def highlightBlock(self, text):
        for pattern, format in self.mapping.items():
            for match in re.finditer(pattern, text):
                start, end = match.span()
                self.setFormat(start, end-start, format)


class Highlight:
    def __init__(self, id, synthax):

        mapping = {}

        for item in synthax.items():
            pattern, format = item
        
            pattern = rf'{pattern}'
            _format = INIT_FORMAT(format)

            mapping[pattern] = _format
        
        widget = WIDGET_ID_SAFE[id]
        synthax = Highlighter(mapping)
        synthax.setDocument(widget.document())
        
        synthax.document()
        

class GET:
    def __init__(self, id):

        '''
        GET ELEMENT

        :param id:

        :method value:
        :method update:
        :method delete:
        :method append:
        :method html:
        :method insert_html:
        :method plain_text:
        :method alignment:
        :method is_default:
        :method is_readonly:
        :method style:
        :method is_checked:
        :method hidden:
        :method hide:
        :method diabled:
        :method disable:
        :method enable:
        :method is_hidden:
        :method is_disabled:
        :method select_all:
        :method copy:
        :method cut:
        :method undo:
        :method redo:
        :method paste:
        :method clear:
        :method add:
        :method remove:
        :method current:
        :method count:
        :method selected_items:
        :method set:
        :method index:
        :method reset:
        :method minimum:
        :method maximum:
        :method is_text_visible:
        :method reject:
        :method accept:
        :method focus:
        :method cursor:
        :method setcursor:
        :method icon:
        :method show:
        :method scrollbar:
        '''
        self.id = id
        self.widget = WIDGET_ID_SAFE[self.id]
        self.widget_type = str(type(self.widget))

    def __repr__(self):
        allowed = [
            [
                GET_ELEM_TYPE_PROGRESS_BAR,
                GET_ELEM_TYPE_SLIDER,
                GET_ELEM_TYPE_DIAL
            ],

            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed[0], self.widget_type):
            return str(self.widget.value())
        elif process_request(allowed[1], self.widget_type):
            return self.widget.toPlainText()
        else:
            return self.widget.text()

    def value(self, value: str | None = None):
        allowed = [
            GET_ELEM_TYPE_PROGRESS_BAR,
            GET_ELEM_TYPE_SLIDER,
            GET_ELEM_TYPE_DIAL
        ]

        if process_request(allowed, self.widget_type):
            self.widget.setValue(int(value))
        else:
            self.widget.setText(str(value))

    def update(self, widget):
        widget_type, widget_items = check_func(widget)

        try:
            layout.indexOf(self.widget)
            LayItOut([layout, self.widget, 'replaceWidget'],
                     widget_type, widget_items)
            layout.removeWidget(self.widget)
            self.widget.deleteLater()
            self.widget = None
        except:
            ...

        # WIDGET_ID_SAFE.pop(self.id)

    def delete(self):
        layout.removeWidget(self.widget)
        self.widget.deleteLater()
        self.widget = None

    def append(self, value: str = ''):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.append(value)

    def html(self, value):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.setHtml(value)

    def insert_html(self, value):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.insertHtml(value)

    def plain_text(self, value: None = None):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            if value != None:
                self.widget.setPlainText(value)
            else:
                return self.widget.toPlainText()

    def alignment(self, value: None = None):
        if value != None:
            make_alignment(self.widget, value)
        else:
            return self.widget.alignment()

    def is_default(self):
        return self.widget.isDefault()

    def is_readonly(self):
        return self.widget.isReadonly()

    def style(self, css: None = None, reset: bool = False):
        if css != None:
            css = get_style_check(css)  # check for css errors
            get_css = self.widget.styleSheet()

            if reset == False:
                if len(get_css) != 0:
                    def temp_handle(css):
                        container = []
                        css = css.split('}')
                        for item in css:
                            item = item.split('{')
                            container.append(item)

                        return container

                    get_css = temp_handle(get_css)
                    css = temp_handle(css)

                    def make_changes(old, new):
                        for item in old:
                            old[old.index(item)] = item.split(':')

                        for item in new:
                            new[new.index(item)] = item.split(':')

                        space = [' ']
                        if space in old:
                            old.remove(space)
                        elif space in new:
                            new.remove(space)

                        for old_item in old:
                            for new_item in new:
                                if old_item[0].strip() == new_item[0].strip() and len(old_item[0].strip()) != 0:
                                    old_item[1] = new_item[1]
                                    new.remove(new_item)

                        for new_item in new:
                            old.append(new_item)

                        for items in old:
                            old[old.index(items)] = ':'.join(items)

                        return '; '.join(old)

                    for already in get_css:
                        for new_css in css:
                            if len(new_css) == 1:
                                if ':' not in already[0] and len(already[0]) != 0:
                                    get_property = already[1].split(';')
                                    new_property = new_css[0].split(';')
                                    already[1] = make_changes(
                                        get_property, new_property)
                                    # already[1] = already[1] + ' ' + new_css[0] + ' '
                            else:
                                if already[0].strip() == new_css[0].strip():
                                    get_property = already[1].split(';')
                                    new_property = new_css[1].split(';')
                                    already[1] = make_changes(
                                        get_property, new_property)
                    else:
                        # print(get_css, new_css)
                        for item in get_css:
                            get_css[get_css.index(item)] = '{'.join(item)
                        else:
                            new_css = '}'.join(get_css)
                            # print(new_css)

                    self.widget.setStyleSheet(new_css)
                else:
                    self.widget.setStyleSheet(css)
            else:
                self.widget.setStyleSheet(css)

    def is_checked(self):
        return self.widget.isChecked()

    def hidden(self, value: bool | None = None):
        if value != None:
            self.widget.setHidden(value)

    def hide(self):
        self.widget.setHidden(True)
    
    def disabled(self, value: bool | None = None):
        if value != None:
            self.widget.setDisabled(value)

    def disable(self):
        self.widget.setDisabled(True)

    def enable(self):
        self.widget.setEnabled(True)

    def is_hidden(self):
        return self.widget.isHidden()

    def is_disabled(self):
            return self.widget.isDisabled()

    def select_all(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.selectAll()

    def copy(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.copy()

    def cut(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.cut()

    def undo(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.undo()

    def redo(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.redo()

    def paste(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.paste()

    def clear(self):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA,
            GET_ELEM_TYPE_LISTWIDGET,
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_TAB_WIDGET,

            GET_ELEM_TYPE_INPUT
        ]

        if process_request(allowed[0:4], self.widget_type):
            self.widget.clear()
        elif process_request(allowed[4], self.widget_type):
            self.widget.setText('')

    def add(self, items):
        allowed = [
            GET_ELEM_TYPE_LISTWIDGET,
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_TAB_WIDGET,
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed[0], self.widget_type):
            add_list_items(self.widget, items)
        elif process_request(allowed[1], self.widget_type):
            if type(items[0]) == sample_list:
                pass
            elif type(items[0]) == sample_string:
                items = [items]
        elif process_request(allowed[2], self.widget_type):
            add_tabs(self.widget, items)
        elif process_request(allowed[3], self.widget_type):
            add_stacks(self.widget, items)

    def remove(self, items):
        allowed = [
            GET_ELEM_TYPE_LISTWIDGET,
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_TAB_WIDGET,
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed[0], self.widget_type):
            self.widget.takeItem(items)
        elif process_request(allowed[1], self.widget_type):
            self.widget.removeItem(items)
        elif process_request(allowed[2], self.widget_type):
            self.widget.removeTab(items)
        elif process_request(allowed[3], self.widget_type):
            def init_remove(widget):
                widget, layout = widget
                if widget != None:
                    self.widget.removeWidget(widget)
            
            
            if type(items) == sample_integer:
                # Handle later
                items = [self.widget.widget(items), None]
                init_remove(items)
            elif type(items[0]) == sample_list:
                for item in items:
                    init_remove(item)
            else:
                init_remove(items)


    def current(self):
        allowed = [
            GET_ELEM_TYPE_LISTWIDGET,
            GET_ELEM_TYPE_SELECT,
            [
                GET_ELEM_TYPE_TAB_WIDGET,
                GET_ELEM_TYPE_STACKED
            ]
        ]

        if process_request(allowed[0], self.widget_type):
            return self.widget.currentRow()
        elif process_request(allowed[1], self.widget_type):
            return self.widget.currentText()
        elif process_request(allowed[2], self.widget_type):
            return self.widget.currentIndex()

    def count(self):
        allowed = [
            GET_ELEM_TYPE_LISTWIDGET,
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_TAB_WIDGET,
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.count()

    def selected_items(self):
        allowed = [
            GET_ELEM_TYPE_LISTWIDGET,
        ]

        if process_request(allowed, self.widget_type):
            selected_items = [items.text()
                              for items in self.widget.selectedItems()]
            return selected_items

    def set(self, value):
        allowed = [
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed, self.widget_type):
            self.widget.setCurrentIndex(value)


    def index(self):
        allowed = [
            GET_ELEM_TYPE_SELECT,
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.currentIndex()

    def reset(self):
        allowed = [
            GET_ELEM_TYPE_PROGRESS_BAR,
        ]

        if process_request(allowed, self.widget_type):
            self.widget.reset()

    def minimum(self):
        allowed = [
            GET_ELEM_TYPE_PROGRESS_BAR,
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.minimum()

    def maximum(self):
        allowed = [
            GET_ELEM_TYPE_PROGRESS_BAR,
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.maximum()

    def is_text_visible(self):
        allowed = [
            GET_ELEM_TYPE_PROGRESS_BAR,
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.isTextVisible()

    def reject(self, result: int = None):
        allowed = [
            GET_ELEM_TYPE_POPUP,
        ]

        if process_request(allowed, self.widget_type):
            self.widget.reject()
            if result != None:
                self.widget.setResult(result)

    def accept(self, result: int = None):
        allowed = [
            GET_ELEM_TYPE_POPUP,
        ]

        if process_request(allowed, self.widget_type):
            self.widget.accept()
            if result != None:
                self.widget.setResult(result)
    
    def focus(self, value: bool = True):
        allowed = [
            GET_ELEM_TYPE_INPUT,
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            self.widget.setFocus()
    

    def has_focus(self, value: bool = True):
        allowed = [
            GET_ELEM_TYPE_INPUT,
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            return self.widget.hasFocus()

    def cursor(self):
        allowed = [
            GET_ELEM_TYPE_INPUT,
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            cursor = self.widget.textCursor()
            position = cursor.position()

            return cursor, position

    
    def setcursor(self, cursor):
        cursor, position = cursor

        cursor.setPosition(position)
        self.widget.setTextCursor(cursor)

    def icon(self, path):
        self.widget.setIcon(QIcon(path))

    def show(self, widget):
        allowed = [
            GET_ELEM_TYPE_STACKED
        ]

        if process_request(allowed, self.widget_type):
            widget, layout = widget
            self.widget.setCurrentWidget(widget)

    def scrollbar(self, scrollbar: None = None, bar_type: str = 'v', id: None = None):
        allowed = [
            GET_ELEM_TYPE_TEXTAREA
        ]

        if process_request(allowed, self.widget_type):
            if scrollbar != None:
                if bar_type == 'v':
                    self.widget.setVerticalScrollBar(scrollbar)
                elif bar_type == 'h':
                    self.widget.setHorizontalScrollBar(scrollbar)
            else:
                if bar_type == 'v':
                    bar =  self.widget.verticalScrollBar()
                elif bar_type == 'h':
                    bar = self.widget.horizontalScrollBar()
                else:
                    bar = None

                if bar != None:
                    return INIT_WIDGET(id, bar)
                else:
                    return bar

class get(GET):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
############################## EASE OF ACCESS #############################
###########################################################################

# ------------------------------------------------------------------------#


class Exit(Button):
    def __init__(self, label: str = 'Exit', icon: None = None, id: None = None,
                 disabled: bool = False, default: bool = False, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), checkable: bool = False, checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type = ELEM_TYPE_BUTTON

        self.rtn = [elem_type, label, Window.quit, icon, id,
                    disabled, default, grid, sizepolicy, checkable, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class exit(Exit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Copy(Button):
    def __init__(self, Target_ID, button_text: str = 'Copy', icon: None = None,
                 id: None = None, disabled: bool = False, default: bool = False,
                 grid: tuple = (None, None), sizePolicy: tuple = (None, None), checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type = elem_type, id = pre_widget_process('copy')
        METHOD_COPY.append(f'{id}:{Target_ID}')

        self.rtn = [elem_type, button_text, action_copy, icon, id,
                    disabled, default, grid, sizePolicy, True, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class copy(Copy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Cut(Button):
    def __init__(self, Target_ID, button_text: str = 'Cut', icon: None = None,
                 id: None = None, disabled: bool = False, default: bool = False,
                 grid: tuple = (None, None), sizePolicy: tuple = (None, None), checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type, id = pre_widget_process('cut')
        METHOD_CUT.append(f'{id}:{Target_ID}')

        self.rtn = [elem_type, button_text, action_cut, icon, id,
                    disabled, default, grid, sizePolicy, True, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class cut(Cut):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Paste(Button):
    def __init__(self, Target_ID, button_text: str = 'Paste', icon: None = None,
                 id: None = None, disabled: bool = False, default: bool = False,
                 grid: tuple = (None, None), sizePolicy: tuple = (None, None), checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type, id = pre_widget_process('paste')
        METHOD_PASTE.append(f'{id}:{Target_ID}')

        self.rtn = [elem_type, button_text, action_paste, icon, id,
                    disabled, default, grid, sizePolicy, True, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class paste(Paste):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Undo(Button):
    def __init__(self, Target_ID, button_text: str = 'Undo', icon: None = None,
                 id: None = None, disabled: bool = False, default: bool = False,
                 grid: tuple = (None, None), sizePolicy: tuple = (None, None), checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type = elem_type, id = pre_widget_process('undo')
        METHOD_UNDO.append(f'{id}:{Target_ID}')

        self.rtn = [elem_type, button_text, action_undo, icon, id,
                    disabled, default, grid, sizePolicy, True, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class undo(Undo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Redo(Button):
    def __init__(self, Target_ID, button_text: str = 'Redo', icon: None = None,
                 id: None = None, disabled: bool = False, default: bool = False,
                 grid: tuple = (None, None), sizePolicy: tuple = (None, None), checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):

        elem_type = elem_type, id = pre_widget_process('redo')
        METHOD_REDO.append(f'{id}:{Target_ID}')

        self.rtn = [elem_type, button_text, action_redo, icon, id,
                    disabled, default, grid, sizePolicy, True, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class redo(Redo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
############################## CONTROL BUTTONS ############################
###########################################################################

# ------------------------------------------------------------------------#


def minimize():
    return GET(MINIMIZE_BUTTON_ID)


def maximize():
    return GET(MAXIMIZE_BUTTON_ID)


def close():
    return GET(CLOSE_BUTTON_ID)


def title():
    return GET(TITLE_TEXT_ID)


# ------------------------------------------------------------------------#

###########################################################################
############################### MANUAL WINDOW #############################
###########################################################################

# ------------------------------------------------------------------------#


# def main():
#     window = Window()
#     Box([[Button('UNAVAILABLE')]])
#     window.run()


# ------------------------------------------------------------------------#

###########################################################################
########################### CLERA STYLING EDITOR ##########################
###########################################################################

# ------------------------------------------------------------------------#

# def key():
#     print(QStyleFactory.keys())

# key()



import os
import os.path as path
import sys
import inspect
import re

from PySide6.QtGui import QAction, QIcon, Qt, QTextCharFormat, QColor,QKeyEvent
from PySide6.QtWidgets import QSizePolicy, QApplication
from PySide6.QtWidgets import QListWidgetItem, QAbstractItemView, QComboBox

from ..icons import *

def start():
    app = QApplication(sys.argv)
    return app


def call(func: None = None, *args, **kwargs):
    if func != None:
        return lambda function: func(*args, **kwargs)

DEFAULT_WINDOW_TITLE = None
DEFAULT_WINDOW_ICON = WINDOW_ICON
DEFAULT_WINDOW_SIZE = (None, None)
DEFAULT_WINDOW_FIXED_SIZE = (None, None)
DEFAULT_WINDOW_STYLE = None
DEFAULT_WINDOW_GEOMETRY = None
DEFAULT_TAB_DISTANCE = 30


WIDGET_ID_SAFE = {}
ID_COUNT_DEFAULT = [0]
TOOLBAR_COUNT = []
LINK_ITEMS_INCLUDED = []

SIZE_POLICY_EXPAND = expand = 'expand'
SIZE_POLICY_FIXED = fixed = 'fixed'

circle = 'circle'
square = 'square'

# GRAB_WIDGET = 'GrabWidget'

DEFAULT_SEPARATOR_MARKER = '---'

ELEM_TYPE_ITEM = 'item'
ELEM_TYPE_SEPARATOR = 'separator'
ELEM_TYPE_COLUMN = 'column'
ELEM_TYPE_GROUP = 'group'
ELEM_TYPE_FIELDSET = 'fieldset'
ELEM_TYPE_EMPTY = 'empty'
ELEM_TYPE_BUTTON = 'button'
ELEM_TYPE_INPUT = 'input'
ELEM_TYPE_TEXT = 'text'
ELEM_TYPE_IMAGE = 'img'
ELEM_TYPE_CHECKBOX = 'checkbox'
ELEM_TYPE_RADIO_BUTTON = 'radio_button'
ELEM_TYPE_TEXTAREA = 'textarea'
ELEM_TYPE_LIST_WIDGET = 'list_widget'
ELEM_TYPE_LIST_ITEM = 'list_item'
ELEM_TYPE_SELECT = 'select'
ELEM_TYPE_OPTION = 'option'
ELEM_TYPE_PROGRESS_BAR = 'progress_bar'
ELEM_TYPE_SLIDER = 'slider'
ELEM_TYPE_DIAL = 'dial'
ELEM_TYPE_POPUP = 'popup'
ELEM_TYPE_TAB = 'tab'
ELEM_TYPE_TAB_WIDGET = 'tab_widget'
ELEM_TYPE_SCROLL_AREA = 'scrollarea'
ELEM_TYPE_STACKED = 'stack'


DEFAULT_VERTICAL_TYPES = ['vertical', 'v']
DEFAULT_HORIZONTAL_TYPES = ['horizontal', 'h']

vertical =DEFAULT_VERTICAL_TYPES[0]
horizontal = DEFAULT_HORIZONTAL_TYPES[0]

SET_HORIZONTAL = 'HORIZONTAL'
SET_VERTICAL = 'VERTICAL'

SET_NORMAL = normal = 'NORMAL'
SET_STATIC = static ='STATIC'

INPUT_TYPE_STANDARD = standard = 'STANDARD'
INPUT_TYPE_PASSWORD = password = 'PASSWORD'
INPUT_TYPE_NO_ECHO = noecho = 'NO_ECHO'


GET_ELEM_TYPE_TEXTAREA = 'QTextEdit'
GET_ELEM_TYPE_LISTWIDGET = 'QListWidget'
GET_ELEM_TYPE_SELECT = 'QComboBox'
GET_ELEM_TYPE_PROGRESS_BAR = 'QProgressBar'
GET_ELEM_TYPE_SLIDER = 'QSlider'
GET_ELEM_TYPE_DIAL = 'QDial'
GET_ELEM_TYPE_TAB_WIDGET = 'QTabWidget'
GET_ELEM_TYPE_POPUP = 'QDialog'
GET_ELEM_TYPE_INPUT = 'QLineEdit'
GET_ELEM_TYPE_STACKED = 'QStackedWidget'
# GET_ELEM_TYPE_SCROLLAREA = 'QScrollArea'

ITEM_SELECTION_MODE_EXTENDED = extended = 'extended'
ITEM_SELECTION_MODE_NO_SELECTION = noselection= 'no_selection'
ITEM_SELECTION_MODE_MULTI = multi = 'multi'
ITEM_SELECTION_MODE_SINGLE = single = 'single'


PRE_INIT_COPY = 'copy'
PRE_INIT_CUT = 'cut'
PRE_INIT_PASTE = 'paste'
PRE_INIT_UNDO = 'undo'
PRE_INIT_REDO = 'redo'


center = 'center'
right = 'right'
left = 'left'
top = 'top'
bottom = 'bottom'
hcenter = 'hcenter'
vcenter = 'vcenter'
justify = 'justify'


CUSTOM_TITLEBAR_HEIGHT = 5

CONTROL_BUTTON_CIRCLE = '10px'
CONTROL_BUTTON_SQUARE = '0'

DEFAULT_BUTTON_STYLE = 'circle'

OPEN_FILE_TYPE_SINGLE = 'single'
OPEN_FILE_TYPE_MULTI = 'multi'

DEFAULT_OPEN_FILE_TYPE = OPEN_FILE_TYPE_SINGLE

FILTER_SPLIT_MARKER = '&'

CONTROL_CLOSE = '⨯'

SYSTEM_PLATFORM_WINDOWS = windows = 'windows'
SYSTEM_PLATFORM_FUSION = fusion = 'fusion'
SYSTEM_PLATFORM_CURVE = curve = 'qtcurve'
SYSTEM_PLATFORM_OXYGEN = oxygen = 'oxygen'
SYSTEM_PLATFORM_BREEZE = breeze = 'breeze'

ICON_IMAGE_ID = '-clera_icon-'
TITLE_TEXT_ID = '-clera_title-'
MINIMIZE_BUTTON_ID = '-clera_minimize_button-'
MAXIMIZE_BUTTON_ID = '-clera_maximize_button-'
CLOSE_BUTTON_ID = '-clera_close_button-'

GRID_IDENTIFIER = '=grid'
BOX_IDENTIFIER = '=box'
# TAB_IDENTIFIER = 'tab('
SCROLL_AREA_IDENTIFIER = '=scrollarea'
# STACK_IDENTIFIER = '=stack'

SYSTEM_OS_PLATFORM = sys.platform.lower()


def FUSION_CONTROLS():
    global CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED

    CONTROL_MINIMIZED = '-'
    CONTROL_RESTORE = '+'
    CONTROL_MAXIMIZED = '+'

    return CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED


def WINDOWS_CONTROLS():
    global CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED

    CONTROL_MINIMIZED = '–'
    CONTROL_RESTORE = '❒'
    CONTROL_MAXIMIZED = '◻'

    return CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED


def CLERA_CONTROLS():
    global CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED

    CONTROL_MINIMIZED = '-'
    CONTROL_RESTORE = '+'
    CONTROL_MAXIMIZED = '+'

    return CONTROL_MINIMIZED, CONTROL_RESTORE, CONTROL_MAXIMIZED


def INIT_WIDGET(ID, widget):
    if ID == None or ID == '':
        ID = ID_COUNT_DEFAULT[0]
        ID_COUNT_DEFAULT[0] += 1

    WIDGET_ID_SAFE[ID] = widget
    return WIDGET_ID_SAFE[ID]


def INIT_ITEM(self, item):
    item_type, label, func, icon, tooltip, statustip, shortcut = item

    if icon != None:
        icon = get_path(icon)

    Action = QAction(QIcon(icon), label, self)
    Action.triggered.connect(func)

    Action.setStatusTip(statustip)
    Action.setToolTip(tooltip)
    if shortcut != None:
        Action.setShortcut(shortcut)

    return Action


def set_size_policy(widget, sizepolicy):
    policy = {
        'fixed'     : QSizePolicy.Fixed,
        'expand'    : QSizePolicy.Expanding
    }

    def init_single_policy(sizepolicy):
        if sizepolicy == SIZE_POLICY_EXPAND or sizepolicy == SIZE_POLICY_FIXED:
            return (policy[sizepolicy], policy[sizepolicy])

    if type(sizepolicy) == sample_tuple:
        if len(sizepolicy) == 2:
            horizontal_policy, vertical_policy = sizepolicy

            if horizontal_policy != None and vertical_policy != None:
                horizontal = horizontal_policy.lower()
                vertical = vertical_policy.lower()

                try:
                    vertical = policy[vertical]
                    horizontal = policy[horizontal]
                except:
                    raise ValueError(sizepolicy)
        elif len(sizepolicy) == 1:
            horizontal, vertical = init_single_policy(sizepolicy)
        else:
            ...
            # raise an error
    elif type(sizepolicy) == sample_string:
        horizontal, vertical = init_single_policy(sizepolicy)
    
    try:
        widget.setSizePolicy(horizontal, vertical)
    except:
        ...
        # raise an error

def set_widget(lyt, grid, widget, grid_pos_x, grid_pos_y):
    gridy, gridx = grid  # vertical (y) and horizontal (x) occupy rule for grid widgets

    if type(lyt) != sample_list and type(lyt) != sample_string:
        if grid_pos_x != None and grid_pos_y != None:
            if gridx != None and gridy != None:
                lyt.addWidget(widget, grid_pos_x, grid_pos_y, gridy, gridx)
            else:
                lyt.addWidget(widget, grid_pos_x, grid_pos_y)
        else:
            lyt.addWidget(widget)
    else:
        if len(lyt) == 2:
            lyt, _type = lyt

            if _type == 'addWidget':
                lyt.addWidget(widget)
            elif _type == 'addPermanentWidget':
                lyt.addPermanentWidget(widget)
            else:
                ...
        elif len(lyt) == 3:
            layout, pre_widget, _type = lyt
            if _type == 'replaceWidget':
                layout.replaceWidget(pre_widget, widget)


def make_alignment(widget, alignments):
    def alignCheck(widget_alignment):
        alignment = widget_alignment.lower()

        alignment_map = {
            center      : Qt.AlignCenter,
            right       : Qt.AlignRight,
            left        : Qt.AlignLeft,
            top         : Qt.AlignTop,
            bottom      : Qt.AlignBottom,
            hcenter     : Qt.AlignHCenter,
            vcenter     : Qt.AlignVCenter,
            justify     : Qt.AlignJustify
        }
        
        try:
            align = alignment_map[alignment]
        except:
            ...
            # raise an error

        return align

    align = alignCheck(alignments)
    widget.setAlignment(align)


def get_parent(identifier):
    inspector = inspect.stack()[2][4]
    result = str(inspector).lower().replace(' ', '')

    if (identifier not in result and f"{ELEM_TYPE_TAB}({identifier.removeprefix('=')}" not in result and 
        ELEM_TYPE_SCROLL_AREA not in result and ELEM_TYPE_STACKED not in result):
        return True
    else:
        return False

def get_filter(value):
    # split_markers = ['), (', ') ,(', ') , (', '),(']
    # for split_marker in split_markers:
    #     if split_marker in value:
    #         value = value.replace(split_marker, FILTER_SPLIT_MARKER)
    # "(Python: *.py) , (Bash: *.sh, *.cmd)"
    # value = value.replace('(', '').replace(')', '')
    
    value = value.split(')')
    value = [item.split('(') for item in value]

    result = [
        item 
            for items in range(len(value)) 
                for item in value[items] 
                    if len(item) != 0 and item.strip() != ','
            ]
    value = FILTER_SPLIT_MARKER.join(result)
    
    def init_filter(filter):
        filter = filter.split(':')
        filter_type = filter[0].strip()
        extensions = filter[1]
        extensions = extensions.split(',')
        filter_extensions = ''
        for extension in extensions:
            filter_extensions += extension.strip() + ' '
        else:
            filter_extensions = filter_extensions[:-1]

        return f'{filter_type} ({filter_extensions})'

    if FILTER_SPLIT_MARKER in value:
        filters = value.split(FILTER_SPLIT_MARKER)
        file_filter = ''
        for filter in filters:
            filter = init_filter(filter)
            file_filter += filter + ' ;; '
        else:
            file_filter = file_filter[:-4]
    else:
        file_filter = init_filter(value)

    return file_filter


def check_func(property):
    # IMPROVE LATER
    if type(property) == type(list()):
        widget_type = property[0]
    elif type(property) == type(str()):
        widget_type = ''
    else:
        widget_type = property()[0]
        property = property()

    return widget_type, property


def process_request(allowed: list = None, widget_type: None = None):
    if type(allowed) == sample_list:
        for values in allowed:
            if values in widget_type:
                return True
        else:
            return False

    elif type(allowed) == sample_string:
        if allowed in widget_type:
            return True
        else:
            return False


def add_list_items(widget, list_items):

    for items in list_items:
        NotRequired, label, icon = items

        if icon != None:
            icon_path = get_path(icon)

        list_widget_item = QListWidgetItem(QIcon(icon_path), label)
        widget.addItem(list_widget_item)


def add_select_option(widget, options):
    for option in options:
        item_type = option[0].lower()

        if item_type == ELEM_TYPE_OPTION:
            option_value = str(option[1])
            option_icon = option[2]

            # print(type(option_value))

            if ELEM_TYPE_INPUT in str(type(option_value)).lower():
                widget.setLineEdit(option_value)
            else:
                widget.addItem(option_value)
                icon_index = widget.count()-1

                if option_icon != None:
                    widget.setItemIcon(
                        icon_index, QIcon(get_path(option_icon)))


def add_tabs(tab_widget, tabs):
    def init_tab(tab):
        elem_type, layout, name, icon = tab

        if elem_type == ELEM_TYPE_TAB:
            if icon != None:
                icon = QIcon(get_path(icon))
            else:
                icon = QIcon(icon)

            tab_widget.addTab(layout, icon, name)

    if type(tabs[0]) == sample_list:
        for tab in tabs:
            init_tab(tab)
    elif type(tabs[0]) == sample_string:
        init_tab(tabs)


def add_stacks(stack_widget, stacks):
    def init_stack(stack):
        widget, layout = stack
        stack_widget.addWidget(widget)
        
    if type(stacks[0]) == sample_list:
        for stack in stacks:
            init_stack(stack)
    else:
        init_stack(stacks)


def get_path(href):
    return path.abspath(href)


# PRE-DEFINED WIDGETS
PRE_DEFINED_WIDGETS = []
PRE_WIDGETS_COUNT = [0]

# All the Pre-Defined Widget Method List
METHOD_COPY = []
METHOD_CUT = []
METHOD_PASTE = []
METHOD_UNDO = []
METHOD_REDO = []


def PRE_INIT(MethodType, MethodItem):
    for item in MethodItem:
        try:
            item = item.split(':')
            pre_id = item[0]
            target_id = item[1]

            pre_widget = WIDGET_ID_SAFE[pre_id]
            target_widget = WIDGET_ID_SAFE[target_id]

            if pre_widget.isChecked():
                if MethodType == PRE_INIT_COPY:
                    pre_widget.clicked.connect(
                        target_widget.copy())
                elif MethodType == PRE_INIT_CUT:
                    pre_widget.clicked.connect(
                        target_widget.cut())
                elif MethodType == PRE_INIT_PASTE:
                    pre_widget.clicked.connect(
                        target_widget.paste())
                elif MethodType == PRE_INIT_UNDO:
                    pre_widget.clicked.connect(
                        target_widget.undo())
                elif MethodType == PRE_INIT_REDO:
                    pre_widget.clicked.connect(
                        target_widget.redo())
                pre_widget.toggle()
        except:
            if pre_widget.isChecked():
                pre_widget.toggle()


def action_copy():
    PRE_INIT(PRE_INIT_COPY, METHOD_COPY)


def action_cut():
    PRE_INIT(PRE_INIT_CUT, METHOD_CUT)


def action_paste():
    PRE_INIT(PRE_INIT_PASTE, METHOD_PASTE)


def action_undo():
    PRE_INIT(PRE_INIT_UNDO, METHOD_UNDO)


def action_redo():
    PRE_INIT(PRE_INIT_REDO, METHOD_REDO)


def pre_widget_process(widget_type):
    id = widget_type + str(PRE_WIDGETS_COUNT[0])
    PRE_WIDGETS_COUNT[0] += 1

    return ELEM_TYPE_BUTTON, id


def sample_func(): ...


sample_string = type(str())

sample_list = type(list())

sample_function = type(sample_func)

sample_integer = type(int())

sample_tuple = type(tuple())

if SYSTEM_OS_PLATFORM == 'linux':
    minimize_padding = ' padding: 1px 7px 2px 7px'
    maximize_padding = 'padding: 1px 6px 2px 6px'
    close_padding = 'padding: 1px 4px 2px 4px'
else:
    minimize_padding = 'padding: 1px 8px 3px 8px'
    maximize_padding = 'padding: 1px 6px 3px 6px'
    close_padding = 'padding: 1px 6px 3px 6px'


def fusion_style(minimize, maximize, close, control_button_style):
    minimize.style(f'''
        button [
            color: rgb(18, 18, 18);
            font-weight: bold;
            {minimize_padding};
            margin: 0 4px 0 5px;
            background: rgb(18, 18, 18);
            border-radius: {control_button_style};
            min-width: 0px;
            height: 17px;
        ]

        button:hover [
            color: white;
        ]
    ''')

    maximize.style(f'''
        button [
            color: rgb(18, 18, 18);
            font-weight: bold;
            {maximize_padding};
            margin-right: 4px;
            background: rgb(18, 18, 18);
            border-radius: {control_button_style};
            min-width: 0px;
            height: 17px;
        ]

        button:hover [
            color: white;
        ]
    ''')

    close.style(f'''
        button [
            color: rgb(18, 18, 18);
            margin-right: 8px;
            font-weight: bold;
            {close_padding};
            background: rgb(56, 109, 209);
            border-radius: {control_button_style};
            min-width: 0px;
            height: 17px;
        ]

        button:hover [
            color: white;
        ]
    ''')


def windows_style(minimize, maximize, close):
    minimize.style(f'''
        button [
            font-weight: bold;
            padding: 5.5px 17px 6.5px 17px;
            background: rgba(65, 63, 63, 0);
            border: 0px solid;
            min-width: 0px;
            color: white;
        ]

        button:hover [
            color: white;
            background: grey;
        ]
    ''')

    maximize.style(f'''
        button [
            font-weight: bold;
            padding: 5.5px 15px 6.5px 15px;
            background: rgba(65, 63, 63, 0);
            border: 0px solid;
            color: white;
            min-width: 0px;

        ]

        button:hover [
            color: white;
            background: grey;
        ]
    ''')

    close.style(f'''
        button [
            font-weight: bold;
            padding: 5.5px 15px;
            background: rgba(65, 63, 63, 0);
            border: 0px solid;
            color: white;
            min-width: 0px;
        ]

        button:hover [
            color: white;
            background: red;
        ]
    ''')


# def clera_style(minimize, maximize, close):
#     minimize.style(f'''
#         button [
#             color: rgb(206, 159, 5);
#             font-weight: bold;
#             margin-left: 4px;
#             padding: 0 7px 2px 7px;
#             background: rgb(206, 159, 5);
#             border-radius: {CONTROL_BUTTON_CIRCLE};
#         ]

#         button:hover [
#             color: rgb(122, 94, 0);
#         ]
#     ''')

#     maximize.style(f'''
#         button [
#             color: rgb(2, 141, 2);
#             font-weight: bold;
#             margin: 0 5px 0 4px;
#             padding: 1px 6px 1px 6px;
#             background: rgb(2, 141, 2);
#             border-radius: {CONTROL_BUTTON_CIRCLE};
#         ]

#         button:hover [
#             color: rgb(0, 41, 0);
#         ]
#     ''')

#     close.style(f'''
#         button [
#             color: rgb(172, 3, 3);
#             margin-left: 8px;
#             padding: 0 4px 1px 4px;
#             font-weight: bold;
#             background: rgb(172, 3, 3);
#             border-radius: {CONTROL_BUTTON_CIRCLE};
#         ]

#         button:hover [
#             color: rgb(61, 0, 0);
#         ]
#     ''')


def append_items(append_list, item):
    append_list.append(item)

COLORS = {
    'red'       : Qt.red,
    'black'     : Qt.black,
    'white'     : Qt.white,
    'gray'      : Qt.gray,
    'green'     : Qt.green,
    'blue'      : Qt.blue,
    'cyan'      : Qt.cyan,
    'magenta'   : Qt.magenta,
    'yellow'    : Qt.yellow,
}

def get_color(value):
    if 'rgb' in value:
        rgb = [0, 0, 0]
        new_color = value.lower()
        new_color = value.replace('(', '').replace(')', '')
        new_color = new_color.replace('rgba', 'rgb')
        color = new_color.removeprefix('rgb') 
        color = color.split(',')

        for _index in range(len(color)):
            if _index <= 2:
                rgb[_index] = int(color[_index])
        else:
            rgb = rgb[0:3]
            red, green, blue = rgb
            color = QColor(red, green, blue)
    else:
        try:
            color = COLORS[value]
        except:
            color = None

    return color


def INIT_FORMAT(elements):
    format = QTextCharFormat()
    KEYS = elements.keys()
    if 'color' in KEYS:
        color = get_color(elements['color'])
        if color != None:
            format.setForeground(color)

    if 'background' in KEYS:
        background = get_color(elements['background'])
        if background != None:
            format.setBackground(background)
            
    return format

def r(list: list = None):
    if list != None:
        return '|'.join([fr'\b{list[i]}\b' for i in range(len(list))])



from .handlers import *

def Column(widgets) -> list: ...

def column(widgets) -> Column: ...

def item(label: any, func: function, icon: str, tooltip: any | str, 
         statustip: any | str, shortcut: str) -> list: ...

def tab(layout, name: str, icon: str) -> list: ...
def option(value: any, icon: str) -> list: ...
def ListItem(label: any, icon: str) -> list: ...
def separator() -> list: ...
def empty() -> list: ...
def fieldset(label: any | str, widgets: list, id: str, orientation: str) -> list: ...
def Group(widgets: list, orientation: str) -> list: ...
def group(widgets: list, orientation: str) -> list: ...

class Button:
    def __init__(self, value: any, func: function, icon: str, id: any, disabled: bool, 
                 default: bool, grid: tuple, sizepolicy: tuple, checkable: bool,
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        '''
        Button Widget

        :param value:
        :param func:
        :param icon:
        :param id:
        :param disabled:
        :param default:
        :param grid:
        :param sizepolicy:
        :param checkable:
        :param checked:
        :param hidden:
        :param focus:
        :param icon_size:
        :param statustip:
        :param tooltip:
        :param shortcut:
        '''
        ...

class button(Button): ...

class Input:
    def __init__(self, placeholder: any, id: any, value: any,
                 type: str, disabled: bool, readonly: bool,
                 maxlength: int, hidden: bool, font: str,
                 fontsize: int, text_changed: function, return_pressed: function,
                 editing_finished: function, text_edited: function,
                 selection_changed: function, sizepolicy: tuple, grid: tuple):
        '''
        Input Widget

        :param placeholder:
        :param id:
        :param value:
        :param type:
        :param diabled:
        :param readonly:
        :param maxlength:
        :param hidden:
        :param font:
        :param fontsize:
        :param text_changed:
        :param return_pressed:
        :param editing_finished:
        :param text_edited:
        :param selection_changed:
        :param sizepolicy:
        :param grid:
        '''
        ...

class input(Input): ...

class Text:
    def __init__(self, value: any, id: any, link: str,
                 hovered: function, clicked: function, buddy: str, alignment: str,
                 wordwrap: bool, grid: tuple,
                 sizepolicy: tuple, hidden: bool):
        '''
        Text Widget

        :param value:
        :param id:
        :param link:
        :param hovered:
        :param clicked:
        :param buddy:
        :param alignment:
        :param wordwrap:
        :param grid:
        :param sizepolicy:
        :param hidden:
        '''

        ...
        
class text(Text): ...

class Image:
    def __init__(self, source: str, id: any, size: int, alignment: str, grid: tuple,
                 sizepolicy: tuple, hidden: bool):
        '''
        Image Widget

        :param source:
        :param id:
        :param size:
        :param alignment:
        :param grid:
        :param sizePolicy:
        :param hidden:
        '''

        ...

class image(Image): ...

class CheckBox:
    def __init__(self, label: any, checked: bool, id: any,
                 state_changed: function, toggled: function, grid: tuple,
                 sizepolicy: tuple):
        '''
        Checkbox Widget

        :param label:
        :param checked:
        :param id:
        :param state_changed:
        :param toggled:
        :param grid:
        :param sizepolicy:
        '''
        ...

class checkbox(CheckBox): ...

class RadioButton:
    def __init__(self, label: any, checked: bool, id: any,
                 toggled: function, grid: tuple,
                 sizepolicy: tuple):
        '''
        RadioButton Widget

        :param label:
        :param checked:
        :param id:
        :param stateChanged:
        :param toggled:
        :param grid:
        :param sizepolicy:
        '''
        ...

class radiobutton(RadioButton): ...

class Textarea:
    def __init__(self, id: any, placeholder: any,
                 hidden: bool, alignment: str, value: any,
                 disabled: bool, readonly: bool, text_changed: function,
                 selection_changed: function, undo_available: function,
                 redo_available: function, maxlength: int, font: str,
                 fontsize: int, sizepolicy: tuple,
                 grid: tuple, tabwidth: int,
                 wordwrap: bool):
        '''
        Textarea Widget

        :param id:
        :param placeholder
        :param hidden:
        :param alignment:
        :param value;
        :param disabled:
        :param readonly:
        :param text_changed:
        :param selection_changed:
        :param undo_available:
        :param redo_available:
        :param maxlegth:
        :param font:
        :param fontsize:
        :param sizepolicy:
        :param grid:
        :param tabwidth:
        :param wordwrap:
        '''
        ...
class textarea(Textarea): ...

class ListWidget:
    def __init__(self, list_items: any, id: any, mode: str, grid: tuple,
                 sizepolicy: tuple, func: function):
        '''
        List Widget

        :param list_items:
        :param id:
        :param mode:
        :param grid:
        :param sizepolicy:
        :param func:
        '''
        ...

class listwidget(ListWidget): ...

class Select:
    def __init__(self, options: list, id: any, placeholder: any, grid: tuple,
                 sizepolicy: tuple, current_text_changed: function, activated: function):
        '''
        Select Widget

        :param options:
        :param id:
        :param placeholder:
        :param grid:
        :param sizepolicy:
        :param current_text_changed:
        :param activated:
        '''

        ...

class select(Select): ...

class ProgressBar:
    def __init__(self, id: any,  min: int, max: int, value: int, orientation: str, 
                 grid: tuple, sizepolicy: tuple, text_visible: bool, inverted: bool,
                 value_changed: function):
        '''
        ProgressBar Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param orientation:
        :param grid:
        :param sizepolicy:
        :param text_visible:
        :param inverted:
        :param value_changed:
        '''
        ...

class progressbar(ProgressBar): ...

class Slider:
    def __init__(self, id: any, min: int, max: int, value: int,
                 step: int, orientation: str, grid: tuple,
                 sizepolicy: tuple, value_changed: function):
        '''
        Slider Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param step:
        :param orientation:
        :param grid:
        :param sizepolicy:
        :param value_changed:
        '''
        ...

class slider(Slider): ...

class Dial:
    def __init__(self, id: any, min: int, max: int, value: int,
                 tick_target: str, tick: bool, wrapping: bool, grid: tuple,
                 sizepolicy: tuple, value_changed: function):
        '''
        Dial Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param tick_target:
        :param tick:
        :param wrapping:
        :param grid:
        :param sizepolicy:
        :param value_changed:
        '''
        ...

class dial(Dial): ...


# from ..widgets import WindowButton, WindowInput, WindowText, WindowRadioButton,
# from ..widgets import WindowCheckbox, WindowImage, WindowTextarea


from ..widgets import *
from .handlers import *


def LayItOut(lyt, elem_type, property, grid_pos_x: None = None, grid_pos_y: None = None):
    if elem_type == ELEM_TYPE_BUTTON:
        [
            NotRequired,
            button_text,
            func, 
            icon,
            id, 
            disabled,
            default,
            grid,
            sizepolicy,
            checkable,
            checked,
            hidden,
            focus,
            icon_size, 
            statustip,
            tooltip,
            shortcut
        ] = property
        widget = window_button(
            lyt, button_text, func, icon, id, disabled, default, sizepolicy, grid, grid_pos_x, grid_pos_y, checkable, checked, hidden, focus, 
            icon_size, statustip, tooltip, shortcut)

    elif elem_type == ELEM_TYPE_INPUT:
        [
            NotRequired,
            placeholder,
            id,
            value,
            type,
            disabled,
            readonly,
            maxlength,
            hidden,
            font,
            fontsize,
            text_changed,
            return_pressed,
            editing_finished,
            text_edited,
            selection_changed,
            sizepolicy,
            grid
        ] = property

        widget = window_input(lyt, placeholder, id, value, type, disabled, readonly, maxlength, hidden,
                              font, fontsize, text_changed, return_pressed, editing_finished, text_edited,
                              selection_changed, sizepolicy, grid, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_TEXT:
        [
            NotRequired,
            text,
            id,
            link,
            hovered,
            clicked,
            buddy,
            alignment,
            wordwrap,
            grid,
            sizepolicy,
            hidden
        ] = property

        widget = window_text(lyt, text, id, link, hovered, clicked, buddy, alignment,
                             wordwrap, sizepolicy, grid, grid_pos_x, grid_pos_y, hidden)

    elif elem_type == ELEM_TYPE_IMAGE:
        [
            NotRequired,
            image,
            id,
            size,
            alignment,
            grid,
            sizepolicy,
            hidden
        ] = property

        widget = window_image(
            lyt, image, id, size, alignment, hidden, sizepolicy, grid, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_CHECKBOX:
        [
            NotRequired,
            name,
            checked,
            id,
            state_changed,
            toggled,
            grid,
            sizepolicy
        ] = property

        widget = window_checkbox(
            lyt, name, checked, id, state_changed, toggled, sizepolicy, grid, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_RADIO_BUTTON:
        [
            NotRequired,
            name,
            checked,
            id,
            toggled,
            grid,
            sizepolicy
        ] = property

        widget = window_radio_button(
            lyt, name, checked, id, toggled, sizepolicy, grid, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_TEXTAREA:
        [
            NotRequired,
            id, placeholder,
            hidden,
            alignment,
            value,
            disabled,
            readonly,
            text_changed,
            selection_changed,
            undo_available,
            redo_available,
            maxlength,
            font,
            fontsize,
            sizepolicy,
            grid,
            tabwidth,
            wordwrap
        ] = property

        widget = window_textarea(lyt, id, placeholder, hidden, alignment, value, disabled, readonly,
                                 text_changed, selection_changed, undo_available, redo_available,
                                 maxlength, font, fontsize, sizepolicy, grid, grid_pos_x, grid_pos_y, 
                                 tabwidth, wordwrap)

    elif elem_type == ELEM_TYPE_LIST_WIDGET:
        [
            NotRequired,
            list_items,
            id,
            mode,
            grid,
            sizepolicy,
            func,
        ] = property
        widget = window_list_widget(lyt, list_items, id, mode, grid, sizepolicy, grid_pos_x,
                                    grid_pos_y, func)

    elif elem_type == ELEM_TYPE_SELECT:
        [
            NotRequired,
            options,
            id,
            placeholder,
            grid,
            sizepolicy,
            current_text_changed,
            activated
        ] = property
        widget = window_select(lyt, options, id, placeholder, grid,
                               sizepolicy, grid_pos_x, grid_pos_y, current_text_changed, activated)

    elif elem_type == ELEM_TYPE_PROGRESS_BAR:
        [
            NotRequired,
            id,
            minimum,
            maximum,
            value,
            orientation,
            grid,
            sizepolicy,
            text_visible,
            inverted,
            value_changed
        ] = property
        widget = window_progress_bar(
            lyt, id, minimum, maximum, value,  orientation, grid,
            sizepolicy, text_visible, inverted, value_changed, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_SLIDER:
        [
            NotRequired,
            id,
            min,
            max,
            value,
            step,
            orientation,
            grid,
            sizepolicy,
            value_changed
        ] = property
        widget = window_slider(lyt, id, min, max, value, step, orientation,
                               grid, sizepolicy, value_changed, grid_pos_x, grid_pos_y)

    elif elem_type == ELEM_TYPE_DIAL:
        [
            NotRequired,
            id,
            min,
            max,
            value,
            tick_target,
            tick,
            wrapping,
            grid,
            sizepolicy,
            value_changed
        ] = property
        widget = window_dial(lyt, id, min, max, value, tick_target, tick, wrapping,
                             grid, sizepolicy, value_changed, grid_pos_x, grid_pos_y)

    try:
        return widget
    except:
        pass



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as readme:
    with codecs.open(os.path.join(here, "CHANGELOG.md"), encoding="utf-8") as changelog:
        LONG_DESCRIPTION = readme.read() + '\n\n\n' + changelog.read()


VERSION = '0.1.0'
DESCRIPTION = 'Write Simple and Quick Python GUI Application'
KEYWORDS = ['gui', 'clera', 'simple', 'simplegui', 'pyside', 'pyside6', 'python', 'easy']


CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
]

# Setting up
setup(
    name="clera",
    version=VERSION,
    author="Erasmus A. Junior",
    author_email="eirasmx@pm.me",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    licence='GNU LGPLv3',
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    python_requires='>=3.7',
    install_requires=['PySide6'],
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    py_modules=['clera'],
    entry_points = {'console_scripts':['clera-editor = clera:editor']},
)



TYPE_MARKER = '%TYPE%'
SEPERATION_MARKER = ':-sep-:'
PRE_MARK = '%<%'
POST_MARK = '%>%'

compr = [
    'button::QPushButton',
    'text:image:img::QLabel',
    'input::QLineEdit',
    'checkbox::QCheckBox',
    'radiobutton::QRadioButton',
    'toolbar::QToolBar',
    'fieldset::QGroupBox',
    'item::QToolButton',
    'textarea::QTextEdit',
    'window::QMainWindow',
    'listwidget::QListWidget',
    'select::QComboBox',
    'progressbar::QProgressBar',
    'slider::QSlider',
    'dial::QDial',
    'tabwidget::QTabWidget',
    'popup::QDialog',
    'group::QGroupBox',
    'scrollbar::QScrollBar',
    'stack::QStackedWidget'
]


def error_check(property):
    if property.count(':') != property.count(';'):
        property = property.split(';')
        for items in property:
            if items.count(':') > 1:
                error = items.split(' ')
                error.remove('')

                cnt = len(error) / 2
                error = error[0:int(cnt)]
                csserr = ' '.join(error).strip()

                raise ValueError(
                    f"'{csserr}' missing a semi-colon")


def console_css(user_css):
    css = []
    user_css = user_css.replace('[', '{').replace(']', '}')
    user_css = user_css.split('}')
    stylesheet = ""

    key_version = []
    make_key_version = {}

    for item in user_css:
        item = item.replace('\n', '').replace('   ', '')
        item = item.split('{')
        if len(item[0]) != 0:
            css.append(item)

    def local(name, property):
        merged_chunk = f'{name.strip()} [{property}]'
        final_chunk = merged_chunk.replace('[', '{').replace(']', ' } ')

        return final_chunk

    for items in css:

        valid = False

        if len(items) == 2:
            name = items[0]
            if ':' in name:
                idx = name.index(':')
                selector = name[idx:]
                name = name[:idx]
            else:
                selector = ''

            chunk_name = name.strip()
            chunk_name = chunk_name.lower()
            property = items[1]

            # checks whether style is missing a semi-colon
            error_check(property)

            for widget_types in compr:
                if chunk_name in widget_types:
                    type_id = compr.index(widget_types)
                    item = compr[type_id].split('::')
                    if chunk_name in item:
                        widget_name = item[-1]
                        stylesheet += local(widget_name + selector, property)

                        valid = True

            if valid != True:
                if '#' in name:
                    name = int(name.replace('#', ''))
                else:
                    name = name.strip()

                if len(selector) != 0:
                    property = selector + SEPERATION_MARKER + property

                # name = name + selector
                # x = [name, property]
                # key_version.append(x)

                KEY = name
                if SEPERATION_MARKER in property:
                    property = property.split(SEPERATION_MARKER)
                    property = f'{TYPE_MARKER}{property[0]}{PRE_MARK + property[1] + POST_MARK}'
                else:
                    property = f'{TYPE_MARKER}{PRE_MARK + property + POST_MARK}'

                property = property.replace(
                    PRE_MARK, '{').replace(POST_MARK, '}')

                if make_key_version.get(KEY):
                    make_key_version[KEY] = make_key_version[KEY] + \
                        ' ' + property
                else:
                    make_key_version[KEY] = property

    for KEY in make_key_version.keys():
        key_version.append([KEY, make_key_version[KEY]])

    return (key_version, stylesheet)


def get_style_check(css):
    css = css.replace('[', '{').replace(']', '}')
    if '{' in css and '}' in css:
        NotRequired, css = console_css(css)
    else:
        errorcheck = css.split(';')
        for items in errorcheck:
            if items.count(':') > 1:
                csserr = items.split(' ')
                csserr = csserr[:2]
                csserr = ' '.join(csserr).strip()
                raise ValueError(f"'{csserr}' missing a semi-colon")

    return css



from .handlers import *
from .widgets import *


def link(href: str) -> None: ...


# ------------------------------------------------------------------------#

###########################################################################
############################## WINDOWS CLASS ##############################
###########################################################################

# ------------------------------------------------------------------------#


class Window:
    def __init__(self, title: str, icon: str, size: tuple, geometry: tuple, 
                 style: str, fixed_size: tuple, frame: bool,
                 movable: bool, top: bool, spacing: int, content_margin: tuple,
                 move: tuple) -> None:
        '''
        :param title:
        :param icon:
        :param size:
        :param geometry:
        :param style:
        :param fixed_size:
        :param frame:
        :param movable:
        :param top:
        :param spacing:
        :param content_margin:
        :param move:
        '''
        ...
        
    def run(self, css: str): ...
    def quit(self): ...
    def update(self, remove_id: any, widget): ...
    def normal(self): ...
    def minimize(self): ...
    def maximize(self): ...
    def details(self) -> dict: ...
    def title(self, value: any | str): ...

    def _move(self, position: tuple, top: bool): ...


# ------------------------------------------------------------------------#

###########################################################################
############################# LAYOUT PROCESSOR ############################
###########################################################################

# ------------------------------------------------------------------------#


def Box(widgets) -> list: ...

def Grid(widgets) -> list: ...
    
class ScrollArea:
    def __init__(self, widgets, id: any, contain: bool):
        '''
        ScrollArea

        :param widgets:
        :param id:
        :param contain:
        '''
        ...

class scrollarea(ScrollArea): ...


class Stack:
    def __init__(self, widgets: list | Box | Grid | TabWidget, id: any,
                current_changed: function, widget_removed: function):
        '''
        Stack Widget

        :param widgets:
        :param id:
        :param current_changed:
        :param widget_removed:
        '''
        ...

    def set(self, index: int): ...

    def show(self, widget): ...

class stack(Stack): ...


# ------------------------------------------------------------------------#

###########################################################################
############################## CORE ELEMENTS ##############################
###########################################################################

# ------------------------------------------------------------------------#


class Menubar:
    def __init__(self, menu_items: list):
        '''
        Menu Bar

        :param menu_items:
        '''
        ...


class menubar(Menubar): ...


class Titlebar:
    def __init__(self, icon: str, title: str, widgets: list, alignment: str, text_color: str,
                 background_color: str, height: any, button_style: str):
        '''
        Titlebar

        :param title:
        :param widgets:
        :param alignment:
        :param text_color:
        :param background_color:
        :param height:
        :param button_style: square or circle
        '''
        ...
        

class titlebar(Titlebar): ...

class Toolbar:
    def __init__(self, name: str, tool_items: list, movable: bool,
                 position: str, id: any, iconsize: tuple,
                 border: bool, orientation: str, newline: bool):
        '''
        Toolbar Element

        :param name:
        :param tool_items:
        :param movable:
        :param position:
        :param id:
        :param iconsize:
        :param border:
        :param orientation:
        :param newline:
        '''
        ...
        

class toolbar(Toolbar): ...

class Statusbar:
    def __init__(self): ...

    def message(self, text: str, time: int | float):
        '''
        Display temporary status message

        :param text:
        :param time: in seconds.
        '''
        ...

    def clear(self): ...

    def add(self, widget, type: str): ...

    def remove(self, id: any): ...


class statusbar(Statusbar): ...


class File:
    def __init__(self): ...

    def open(self, caption: str, filter: str, directory: str, type: str) -> list:
        '''
        Open File Dialog

        :param caption:
        :param filter:
        :param directory:
        :param type:
        '''
        ...

    def save(self, filter: str = '(All Files: *)'):
        '''
        Save File Dialog

        :param filter:
        '''
        ...

class file(File): ...

class Folder:
    def __init__(self, caption: str, directory: str) -> str:
        '''
        Folder Dialog

        :param caption:
        :param directory:
        '''
        ...

class folder(Folder): ...

class Popup:
    def __init__(self, title: any, widgets: list, size: tuple, id: any,
                 modal: bool, frame: bool, lock: bool) -> list:
        '''
        Popup Window

        :param title:
        :param widgets:
        :param size:
        :param id:
        :param modal:
        :param frame:
        :param lock:
        '''
        ...

    def result(self): ...
    # has a return value

class popup(Popup): ...


class TabWidget:
    def __init__(self, tabs: tab | list, id: any, movable: bool, closable: bool,
                 close_requested: function, clicked: function) -> list:
        '''
        Tab Widget

        :param tabs:
        :param id:
        :param movable:
        :param closable:
        :param close_requested:
        :param clicked:
        '''
        ...

class tabwidget(TabWidget): ...

class Highlight:
    def __init__(self, id: any, synthax: dict): ...


class GET:
    def __init__(self, id: any):

        '''
        GET ELEMENT

        :param id:

        :method value:
        :method update:
        :method delete:
        :method append:
        :method html:
        :method insert_html:
        :method plain_text:
        :method alignment:
        :method is_default:
        :method is_readonly:
        :method style:
        :method is_checked:
        :method hidden:
        :method hide:
        :method diabled:
        :method disable:
        :method enable:
        :method is_hidden:
        :method is_disabled:
        :method select_all:
        :method copy:
        :method cut:
        :method undo:
        :method redo:
        :method paste:
        :method clear:
        :method add:
        :method remove:
        :method current:
        :method count:
        :method selected_items:
        :method set:
        :method index:
        :method reset:
        :method minimum:
        :method maximum:
        :method is_text_visible:
        :method reject:
        :method accept:
        :method focus:
        :method cursor:
        :method setcursor:
        :method icon:
        :method show:
        :method scrollbar:
        '''

        ...
        
    def __repr__(self) -> str: ...

    def value(self, value: str): ...
    def update(self, widget): ...
    def delete(self): ...
    def append(self, value: str): ...
    def html(self, value: str): ...
    def insert_html(self, value: str): ...
    def plain_text(self, value: str | int) -> str: ...
    def alignment(self, value: str) -> str: ...
    def is_default(self) -> bool: ...
    def is_readonly(self) -> bool: ...
    def style(self, css: str, reset: bool): ...
    def is_checked(self) -> bool: ...
    def hidden(self, value: bool): ...
    def hide(self): ...
    def disabled(self, value: bool): ...
    def disable(self): ...
    def enable(self): ...
    def is_hidden(self) -> bool: ...
    def is_disabled(self) -> bool: ...
    def select_all(self): ...
    def copy(self): ...
    def cut(self): ...
    def undo(self): ...
    def redo(self): ...
    def paste(self): ...
    def clear(self): ...
    def add(self, items): ...
    def remove(self, items): ...
    def current(self) -> str | int: ...
    def count(self) -> int: ...
    def selected_items(self) -> list: ...
    def set(self, value): ...
    def index(self) -> int: ...
    def reset(self): ...
    def minimum(self) -> int: ...
    def maximum(self) -> int: ...
    def is_text_visible(self) -> bool: ...
    def reject(self, result: int): ...
    def accept(self, result: int): ...
    def focus(self, value: bool): ...
    def has_focus(self, value: bool) -> bool: ...
    def cursor(self) -> tuple: ...
    def setcursor(self, cursor: tuple): ...
    def icon(self, path: str): ...
    def show(self, widget): ...
    def scrollbar(self, scrollbar, bar_type: str, id: any): ...

class get(GET): ...

# ------------------------------------------------------------------------#

###########################################################################
############################## EASE OF ACCESS #############################
###########################################################################

# ------------------------------------------------------------------------#


class Exit(Button):
    def __init__(self, label: str, icon: str, id: any, disabled: bool, default: bool,
                 grid: tuple, sizePolicy: tuple, checkable: bool, checked: bool, hidden: bool, 
                 focus: bool, icon_size: int, statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class exit(Exit): ...

class Copy(Button):
    def __init__(self, Target_ID: any, button_text: str, icon: str, id: any, 
                 disabled: bool, default: bool, grid: tuple, sizePolicy: tuple, 
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class copy(Copy): ...

class Cut(Button):
    def __init__(self, Target_ID: any, button_text: str, icon: str, id: any, 
                 disabled: bool, default: bool, grid: tuple, sizePolicy: tuple, 
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class cut(Cut): ...

class Paste(Button):
    def __init__(self, Target_ID: any, button_text: str, icon: str, id: any, 
                 disabled: bool, default: bool, grid: tuple, sizePolicy: tuple, 
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class paste(Paste): ...

class Undo(Button):
    def __init__(self, Target_ID: any, button_text: str, icon: str, id: any, 
                 disabled: bool, default: bool, grid: tuple, sizePolicy: tuple, 
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class undo(Undo): ...

class Redo(Button):
    def __init__(self, Target_ID: any, button_text: str, icon: str, id: any, 
                 disabled: bool, default: bool, grid: tuple, sizePolicy: tuple, 
                 checked: bool, hidden: bool, focus: bool, icon_size: int, 
                 statustip: any | str, tooltip: any | str, shortcut: str):
        ...

class redo(Redo): ...

# ------------------------------------------------------------------------#

###########################################################################
############################## CONTROL BUTTONS ############################
###########################################################################

# ------------------------------------------------------------------------#


def minimize() -> GET: ...
def maximize() -> GET: ...
def close() -> GET: ...
def title() -> GET: ...



from .utils import *


from PySide6.QtGui import QIcon, QFont, QPixmap, QTextCursor
from PySide6.QtWidgets import QPushButton, QLabel, QLineEdit, QRadioButton, QCheckBox, QTextEdit, QListWidget
from PySide6.QtWidgets import QComboBox, QProgressBar, QSlider, QDial, QWidget, QScrollArea
from PySide6.QtCore import QSize

def Column(widgets):
    type = ELEM_TYPE_COLUMN
    return [type, widgets]


def column(*args, **kwargs):
    return Column(*args, **kwargs)


def item(label: any = '', func: None = None, icon: None = None,
         tooltip: None = None, statustip: None = None, shortcut: None = None):
    type = ELEM_TYPE_ITEM
    return [type, label, func, icon, tooltip, statustip, shortcut]


def tab(layout: None = None, name: None = None, icon: None = None):
    type = ELEM_TYPE_TAB
    NotRequired, layout = check_func(layout)
    return [type, layout[0], name, icon]


def option(value: any = '', icon: None = None):
    type = ELEM_TYPE_OPTION
    return [type, value, icon]


def ListItem(label: any = '', icon: None = None):
    type = ELEM_TYPE_LIST_ITEM
    return [type, label, icon]


def separator():
    type = ELEM_TYPE_SEPARATOR
    return [type]


def empty():
    type = ELEM_TYPE_EMPTY
    return [type]


def fieldset(label: str = '', widgets: list = None, id: None = None, orientation: str = SET_VERTICAL):
    elem_type = ELEM_TYPE_FIELDSET
    return [elem_type, label, widgets, id, orientation]


def Group(widgets: list = None, orientation: str = SET_VERTICAL):
    elem_type = ELEM_TYPE_GROUP
    return [elem_type, widgets, orientation]


def group(*args, **kwargs):
    return Group(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
#################################  BUTTON  ################################
###########################################################################

# ---------------ToolBarArea---------------------------------------------------------#


def window_button(lyt, button_text, func, icon,
                  ID, disabled, default, sizepolicy, grid, grid_pos_x, grid_pos_y, checkable,
                  checked, hidden, focus, icon_size, statustip, tooltip, shortcut):

    button = INIT_WIDGET(ID, QPushButton(button_text))
    set_widget(lyt, grid, button, grid_pos_x, grid_pos_y)
    set_size_policy(button, sizepolicy)

    button.setStatusTip(statustip)
    button.setToolTip(tooltip)
    if shortcut != None:
        button.setShortcut(shortcut)

    button.setIcon(QIcon(icon))
    button.setIconSize(QSize(icon_size, icon_size))
    button.setDisabled(disabled)
    button.setDefault(default)
    button.setCheckable(checkable)
    button.setChecked(checked)
    button.clicked.connect(func)
    button.setHidden(hidden)

    # button.setFocus(Qt.NoFocusReason)
    if focus == False:
        button.setFocusPolicy(Qt.NoFocus)

    return button


class Button:
    def __init__(self, value: any = '', func: None = None, icon: str = None, id: any = None,
                 disabled: bool = False, default: bool = False, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), checkable: bool = False,
                 checked: bool = False, hidden: bool = False, focus: bool = True, 
                 icon_size: int = 20, statustip: str = None, tooltip: str = None, shortcut: str = None):
        '''
        Button Widget

        :param value:
        :param func:
        :param icon:
        :param id:
        :param disabled:
        :param default:
        :param grid:
        :param sizepolicy:
        :param checkable:
        :param checked:
        :param hidden:
        :param focus:
        :param icon_size:
        :param statustip:
        :param tooltip:
        :param shortcut:
        '''

        elem_type = ELEM_TYPE_BUTTON
        self.rtn = [elem_type, str(value), func, icon, id,
                    disabled, default, grid, sizepolicy, checkable, checked, hidden, focus, icon_size, statustip, tooltip, shortcut]

    def __call__(self):
        return self.rtn


class button(Button):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
#################################  INPUT  #################################
###########################################################################

# ------------------------------------------------------------------------#


def window_input(lyt, placeholder, ID, value, type, disabled,
                 readonly, maxlength, hidden, font, fontsize, text_changed, return_pressed,
                 editing_finished, text_edited, selection_changed, sizepolicy, grid,
                 grid_pos_x, grid_pos_y,):

    input = INIT_WIDGET(ID, QLineEdit())

    set_widget(lyt, grid, input, grid_pos_x, grid_pos_y)
    set_size_policy(input, sizepolicy)

    if fontsize != None:
        input.setFont(QFont(font, fontsize))
    else:
        input.setFont(QFont(font))

    if maxlength != None:
        input.setMaxLength(maxlength)

    input.setHidden(hidden)
    input.setDisabled(disabled)
    input.setReadOnly(readonly)
    input.setText(value)
    input.setPlaceholderText(placeholder)
    input.textChanged.connect(text_changed)
    input.returnPressed.connect(return_pressed)
    input.editingFinished.connect(editing_finished)
    input.textEdited.connect(text_edited)
    input.selectionChanged.connect(selection_changed)

    input_type = type.upper()

    if input_type == INPUT_TYPE_PASSWORD:
        input.setEchoMode(QLineEdit.Password)
    elif input_type == INPUT_TYPE_STANDARD:
        input.setEchoMode(QLineEdit.Normal)
    elif input_type == INPUT_TYPE_NO_ECHO:
        input.setEchoMode(QLineEdit.NoEcho)
    else:
        raise ValueError(type)

    return input


class Input:
    def __init__(self, placeholder: any = None, id: any = None, value: any = None,
                 type: str = INPUT_TYPE_STANDARD, disabled: bool = False, readonly: bool = False,
                 maxlength: int | None = None, hidden: bool = False, font: str | None = None,
                 fontsize: int | None = None, text_changed: None = None, return_pressed: None = None,
                 editing_finished: None = None, text_edited: None = None,
                 selection_changed: None = None, sizepolicy: tuple = (None, None), grid: tuple = (None, None)):
        '''
        Input Widget

        :param placeholder:
        :param id:
        :param value:
        :param type:
        :param diabled:
        :param readonly:
        :param maxlength:
        :param hidden:
        :param font:
        :param fontsize:
        :param text_changed:
        :param return_pressed:
        :param editing_finished:
        :param text_edited:
        :param selection_changed:
        :param sizepolicy:
        :param grid:
        '''

        elem_type = ELEM_TYPE_INPUT
        self.rtn = [elem_type, placeholder, id, value, type, disabled,
                    readonly, maxlength, hidden, font, fontsize, text_changed, return_pressed,
                    editing_finished, text_edited, selection_changed, sizepolicy, grid]

    def __call__(self):
        return self.rtn


class input(Input):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
##################################  TEXT  #################################
###########################################################################

# ------------------------------------------------------------------------#


def window_text(lyt, text, ID, link, hovered, clicked, buddy, alignment, word_wrap,
                sizepolicy, grid, grid_pos_x, grid_pos_y, hidden):

    label = INIT_WIDGET(ID, QLabel(text))

    set_widget(lyt, grid, label, grid_pos_x, grid_pos_y)
    set_size_policy(label, sizepolicy)

    if link != None:
        anchor = f'<a href="{link}">{text}</a>'
        label.setText(anchor)
        label.linkActivated.connect(clicked)
        label.linkHovered.connect(hovered)

    if buddy != None:
        label.setBuddy(WIDGET_ID_SAFE[buddy])

    if alignment != None:
        make_alignment(label, alignment)

    label.setWordWrap(word_wrap)
    label.setHidden(hidden)

    return label


class Text:
    def __init__(self, value: any = '', id: any = None, link: str = None,
                 hovered: None = None, clicked: None = None, buddy: str = None, alignment: str = None,
                 wordwrap: bool = False, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), hidden: bool = False):
        '''
        Text Widget

        :param value:
        :param id:
        :param link:
        :param hovered:
        :param clicked:
        :param buddy:
        :param alignment:
        :param wordwrap:
        :param grid:
        :param sizepolicy:
        :param hidden:
        '''

        elem_type = ELEM_TYPE_TEXT
        self.rtn = [elem_type, str(value), id, link,
                    hovered, clicked, buddy, alignment, wordwrap, grid, sizepolicy, hidden]

    def __call__(self):
        return self.rtn


class text(Text):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
#################################  IMAGE  #################################
###########################################################################

# ------------------------------------------------------------------------#


def window_image(lyt, image, ID, size, alignment, hidden, sizepolicy, grid, grid_pos_x, grid_pos_y,):

    img = INIT_WIDGET(ID, QLabel())

    set_widget(lyt, grid, img, grid_pos_x, grid_pos_y)
    set_size_policy(img, sizepolicy)

    if image != None:
        pixmap = QPixmap(get_path(image))

        if type(size) == sample_integer:
            pixmap = pixmap.scaledToWidth(size)
        elif type(size) == sample_tuple:
            if len(size) == 2:
                height, width = size
                pixmap = pixmap.scaled(QSize(height, width))

        img.setPixmap(pixmap)

        if alignment != None:
            make_alignment(img, alignment)

        img.setHidden(hidden)

    return img


class Image:
    def __init__(self, source: str = None, id: any = None, size: int = None, alignment: str = None, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), hidden: bool = False):
        '''
        Image Widget

        :param source:
        :param id:
        :param size:
        :param alignment:
        :param grid:
        :param sizePolicy:
        :param hidden:
        '''

        elem_type = ELEM_TYPE_IMAGE
        self.rtn = [elem_type, source, id, size,
                    alignment, grid, sizepolicy, hidden]

    def __call__(self):
        return self.rtn


class image(Image):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
################################  CHECKBOX  ###############################
###########################################################################

# ------------------------------------------------------------------------#


def window_checkbox(lyt, name, checked, ID, state_changed, toggled, sizepolicy,
                    grid, grid_pos_x, grid_pos_y,):

    checkbox = INIT_WIDGET(ID, QCheckBox(name))

    set_widget(lyt, grid, checkbox, grid_pos_x, grid_pos_y)
    set_size_policy(checkbox, sizepolicy)

    checkbox.setChecked(checked)
    checkbox.stateChanged.connect(state_changed)
    checkbox.toggled.connect(toggled)

    return checkbox


class CheckBox:
    def __init__(self, label: any = '', checked: bool = False, id: any = None,
                 state_changed: None = None, toggled: None = None, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None)):
        '''
        Checkbox Widget

        :param label:
        :param checked:
        :param id:
        :param state_changed:
        :param toggled:
        :param grid:
        :param sizepolicy:
        '''

        elem_type = ELEM_TYPE_CHECKBOX
        self.rtn = [elem_type, label, checked, id,
                    state_changed, toggled, grid, sizepolicy]

    def __call__(self):
        return self.rtn


class checkbox(CheckBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
##############################  RADIOBUTTON  ##############################
###########################################################################

# ------------------------------------------------------------------------#


def window_radio_button(lyt, name, checked, ID, toggled, sizepolicy, grid,
                        grid_pos_x, grid_pos_y):

    radio_button = INIT_WIDGET(ID, QRadioButton(name))

    set_widget(lyt, grid, radio_button, grid_pos_x, grid_pos_y)
    set_size_policy(radio_button, sizepolicy)

    radio_button.setChecked(checked)
    radio_button.toggled.connect(toggled)

    return radio_button


class RadioButton:
    def __init__(self, label: any = '', checked: bool = False, id: any = None,
                 toggled: None = None, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None)):
        '''
        RadioButton Widget

        :param label:
        :param checked:
        :param id:
        :param stateChanged:
        :param toggled:
        :param grid:
        :param sizepolicy:
        '''

        elem_type = ELEM_TYPE_RADIO_BUTTON
        self.rtn = [elem_type, label, checked, id, toggled, grid, sizepolicy]

    def __call__(self):
        return self.rtn


class radiobutton(RadioButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
################################  TEXTAREA  ###############################
###########################################################################

# ------------------------------------------------------------------------#


def window_textarea(lyt, ID, placeholder, hidden, alignment, value, disabled, readonly, text_changed,
                    selection_changed, undo_available, redo_available, maxlength, font, fontsize,
                    sizepolicy, grid, grid_pos_x, grid_pos_y, tabwidth, wordwrap):

    textarea = INIT_WIDGET(ID, QTextEdit())

    set_widget(lyt, grid, textarea, grid_pos_x, grid_pos_y)
    set_size_policy(textarea, sizepolicy)
    
    #  set word wrap for textedit
    if not wordwrap:
        textarea.setLineWrapMode(textarea.LineWrapMode.NoWrap)

    textarea.setHidden(hidden)
    textarea.setDisabled(disabled)
    textarea.setReadOnly(readonly)
    textarea.setText(value)
    textarea.setPlaceholderText(placeholder)

    textarea.textChanged.connect(text_changed)
    textarea.selectionChanged.connect(selection_changed)
    textarea.undoAvailable.connect(undo_available)
    textarea.redoAvailable.connect(redo_available)

    textarea.setTabStopDistance(tabwidth)

    if fontsize != None:
        textarea.setFont(QFont(font, fontsize))
    else:
        textarea.setFont(QFont(font))

    if maxlength != None:
        textarea.setMaxLength(maxlength)

    if alignment != None:
        make_alignment(textarea, alignment)

    return textarea


class Textarea:
    def __init__(self, id: any = None, placeholder: any = None,
                 hidden: bool = False, alignment: str = None, value: any = None,
                 disabled: bool = False, readonly: bool = False, text_changed: None = None,
                 selection_changed: None = None, undo_available: None = None,
                 redo_available: None = None, maxlength: int | None = None, font: str | None = None,
                 fontsize: int | None = None, sizepolicy: tuple = (None, None),
                 grid: tuple = (None, None), tabwidth: int = DEFAULT_TAB_DISTANCE,
                 wordwrap: bool = True):
        '''
        Textarea Widget

        :param id:
        :param placeholder
        :param hidden:
        :param alignment:
        :param value;
        :param disabled:
        :param readonly:
        :param text_changed:
        :param selection_changed:
        :param undo_available:
        :param redo_available:
        :param maxlegth:
        :param font:
        :param fontsize:
        :param sizepolicy:
        :param grid:
        :param tabWidth:
        :param wordwrap:
        '''
        elem_type = ELEM_TYPE_TEXTAREA
        self.rtn = [elem_type, id, placeholder, hidden, alignment, value, disabled, readonly, text_changed,
                    selection_changed, undo_available, redo_available, maxlength, font, fontsize, sizepolicy, grid,
                    tabwidth, wordwrap]

    def __call__(self):
        return self.rtn


class textarea(Textarea):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
###############################  LISTWIDGET  ##############################
###########################################################################

# ------------------------------------------------------------------------#


def window_list_widget(lyt, list_items, ID, mode, grid, sizepolicy, grid_pos_x, grid_pos_y, func):
    list_widget = INIT_WIDGET(ID, QListWidget())

    set_widget(lyt, grid, list_widget, grid_pos_x, grid_pos_y)
    set_size_policy(list_widget, sizepolicy)

    selection_modes = [
        [ITEM_SELECTION_MODE_NO_SELECTION, QAbstractItemView.NoSelection],
        [ITEM_SELECTION_MODE_SINGLE, QAbstractItemView.SingleSelection],
        [ITEM_SELECTION_MODE_MULTI, QAbstractItemView.MultiSelection],
        [ITEM_SELECTION_MODE_EXTENDED, QAbstractItemView.ExtendedSelection]
    ]

    # def pass_value(item):
    #     func(item.text())

    # list_widget.currentItemChanged.connect(pass_value)

    # if func != None:
    #     list_widget.itemClicked.connect(pass_value)

    # signals and hidden setup
    # create list widget id system.

    for items in selection_modes:

        new_mode = mode.lower()

        if new_mode in items:
            list_mode = items[1]
            list_widget.setSelectionMode(list_mode)
            break
    else:
        raise ItemSelectionModeError(f"{mode} is not a valid mode")

    add_list_items(list_widget, list_items)


class ListWidget:
    def __init__(self, list_items: any = None, id: any = None, mode: str = ITEM_SELECTION_MODE_SINGLE, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), func: None = None):
        '''
        List Widget

        :param list_items:
        :param id:
        :param mode:
        :param grid:
        :param sizepolicy:
        :param func:
        '''

        elem_type = ELEM_TYPE_LIST_WIDGET
        self.rtn = [elem_type, list_items, id, mode, grid, sizepolicy, func]

    def __call__(self):
        return self.rtn


class listwidget(ListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
################################   SELECT   ###############################
###########################################################################

# ------------------------------------------------------------------------#


def window_select(lyt, options, ID, placeholder, grid, sizepolicy, grid_pos_x, grid_pos_y, current_text_changed, activated):
    select = INIT_WIDGET(ID, QComboBox())

    set_widget(lyt, grid, select, grid_pos_x, grid_pos_y)
    set_size_policy(select, sizepolicy)

    select.setPlaceholderText(placeholder)

    if options != None:
        if type(options[0]) == sample_string:
            options = [options]
        elif type(options[0]) == sample_list:
            pass

        add_select_option(select, options)

    select.currentTextChanged.connect(current_text_changed)
    select.activated.connect(activated)


class Select:
    def __init__(self, options: any = None, id: any = None, placeholder: any = None, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), current_text_changed: None = None, activated: None = None):
        '''
        Select Widget

        :param options:
        :param id:
        :param placeholder:
        :param grid:
        :param sizepolicy:
        :param current_text_changed:
        :param activated:
        '''

        elem_type = ELEM_TYPE_SELECT

        self.rtn = [elem_type, options, id, placeholder, grid,
                    sizepolicy, current_text_changed, activated]

    def __call__(self):
        return self.rtn


class select(Select):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
##############################   PROGRESSBAR   ############################
###########################################################################

# ------------------------------------------------------------------------#


def window_progress_bar(lyt, ID, minimum, maximum, value, orientation, grid,
                        sizepolicy, text_visible, inverted, value_changed, grid_pos_x, grid_pos_y):
    progress_bar = INIT_WIDGET(ID, QProgressBar())

    set_widget(lyt, grid, progress_bar, grid_pos_x, grid_pos_y)
    set_size_policy(progress_bar, sizepolicy)

    progress_bar.setRange(minimum, maximum)

    if value != None:
        progress_bar.setValue(value)

    progress_bar.setInvertedAppearance(inverted)

    progress_bar.setTextVisible(text_visible)

    if orientation.lower() in DEFAULT_VERTICAL_TYPES:
        progress_bar.setOrientation(Qt.Vertical)
    elif orientation.lower() in DEFAULT_HORIZONTAL_TYPES:
        progress_bar.setOrientation(Qt.Horizontal)
    else:
        raise ValueError(orientation)

    # progress_bar.setTextDirection(QProgressBar.BottomToTop)

    progress_bar.valueChanged.connect(value_changed)


class ProgressBar:
    def __init__(self, id: any = None,  min: int = 0, max: int = 0, value: int = None,
                 orientation: str = SET_HORIZONTAL, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), text_visible: bool = True, inverted: bool = False, value_changed: None = None):
        '''
        ProgressBar Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param orientation:
        :param grid:
        :param sizepolicy:
        :param text_visible:
        :param inverted:
        :param value_changed:
        '''

        elem_type = ELEM_TYPE_PROGRESS_BAR

        self.rtn = [elem_type, id,  min, max, value, orientation, grid,
                    sizepolicy, text_visible, inverted, value_changed]

    def __call__(self):
        return self.rtn


class progressbar(ProgressBar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
#################################   SLIDER   ##############################
###########################################################################

# ------------------------------------------------------------------------#


def window_slider(lyt, ID, min, max, value, step, orientation, grid,
                  sizepolicy, value_changed, grid_pos_x, grid_pos_y):

    if orientation.lower() in DEFAULT_VERTICAL_TYPES:
        setOrientation = Qt.Vertical
    elif orientation.lower() in DEFAULT_HORIZONTAL_TYPES:
        setOrientation = Qt.Horizontal
    else:
        raise ValueError(orientation)

    slider = INIT_WIDGET(ID, QSlider(setOrientation))

    set_widget(lyt, grid, slider, grid_pos_x, grid_pos_y)
    set_size_policy(slider, sizepolicy)

    slider.setMinimum(min)
    slider.setMaximum(max)

    if value != None:
        slider.setValue(value)

    if step != None:
        slider.setSingleStep(step)

    slider.valueChanged.connect(value_changed)


class Slider:
    def __init__(self, id: any = None, min: int = 0, max: int = 0, value: int = None,
                 step: None = None, orientation: str = SET_HORIZONTAL, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), value_changed: None = None):
        '''
        Slider Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param step:
        :param orientation:
        :param grid:
        :param sizepolicy:
        :param value_changed:
        '''

        elem_type = ELEM_TYPE_SLIDER

        self.rtn = [elem_type, id,  min, max, value,
                    step, orientation, grid, sizepolicy, value_changed]

    def __call__(self):
        return self.rtn


class slider(Slider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# ------------------------------------------------------------------------#

###########################################################################
##################################   DIAL   ###############################
###########################################################################

# ------------------------------------------------------------------------#


def window_dial(lyt, ID, min, max, value, tick_target, tick, wrapping, grid,
                sizepolicy, value_changed, grid_pos_x, grid_pos_y):

    dial = INIT_WIDGET(ID, QDial())

    set_widget(lyt, grid, dial, grid_pos_x, grid_pos_y)
    set_size_policy(dial, sizepolicy)

    dial.setMinimum(min)
    dial.setMaximum(max)

    if value != None:
        dial.setValue(value)

    if tick_target != None:
        dial.setNotchTarget(tick_target)

    dial.setNotchesVisible(tick)
    dial.setWrapping(wrapping)

    dial.valueChanged.connect(value_changed)


class Dial:
    def __init__(self, id: any = None, min: int = 0, max: int = 0, value: int = None,
                 tick_target: None = None, tick: bool = False, wrapping: bool = False, grid: tuple = (None, None),
                 sizepolicy: tuple = (None, None), value_changed: None = None):
        '''
        Dial Widget

        :param id:
        :param min:
        :param max:
        :param value:
        :param tick_target:
        :param tick:
        :param wrapping:
        :param grid:
        :param sizepolicy:
        :param value_changed:
        '''

        elem_type = ELEM_TYPE_DIAL

        self.rtn = [elem_type, id,  min, max, value,
                    tick_target, tick, wrapping, grid, sizepolicy, value_changed]

    def __call__(self):
        return self.rtn


class dial(Dial):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
